#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
"""
.. Created on Wed Oct 16 14:20:49 2019

.. codeauthor: Michael J. Hayford
"""

import math
import numpy as np
from copy import deepcopy
from pathlib import Path

from rayoptics.gui.util import GUIHandle, bbox_from_poly, fit_data_range

from rayoptics.optical.model_constants import ht, slp
from rayoptics.optical.model_constants import pwr, tau, indx, rmd
from rayoptics.util.rgb2mpl import rgb2mpl
from rayoptics.util.misc_math import (normalize, distance_sqr_2d,
                                      projected_point_on_radial_line)
from rayoptics.util.line_intersection import get_intersect
from rayoptics.util import misc_math
from rayoptics.util import colors

from rayoptics.optical.elements import (create_thinlens, create_mirror,
                                        create_lens, create_from_file,
                                        AirGap)


def light_or_dark(is_dark=True):
    accent = colors.accent_colors(is_dark)
    fb = colors.foreground_background(is_dark)
    rgb = {
        'node': fb['foreground'],
        'edge': fb['foreground'],
        # 'node': accent['cyan'],
        # 'edge': accent['cyan'],
        'slide': accent['blue'],
        'object_image': accent['magenta'],
        'stop': accent['magenta'],
        'conj_line': accent['orange'],
        'shift': accent['orange'],
        'barrel': accent['green'],
        }
    return {**rgb, **fb}


def create_parax_design_commands(fig):
    cmds = []
    dgm = fig.diagram
    # initialize dgm with a Select command
    dgm.register_commands((), figure=fig)
    # Select an existing point
    cmds.append(('Select', (dgm.register_commands, (), {})))
    # Add thin lens
    cmds.append(('Add Thin Lens',
                 (dgm.register_add_replace_element, (),
                  {'node_init': create_thinlens,
                   'factory': create_thinlens,
                   'interact_mode': 'transmit'})))
    # Add lens
    cmds.append(('Add Lens', (dgm.register_add_replace_element, (),
                              {'node_init': create_thinlens,
                               'factory': create_lens,
                               'interact_mode': 'transmit'})))
    # Add mirror
    cmds.append(('Add Mirror',
                 (dgm.register_add_replace_element, (),
                  {'node_init': create_mirror,
                   'factory': create_mirror,
                   'interact_mode': 'reflect'})))
    # replace with file
    pth = Path(__file__).resolve()
    root_pos = pth.parts.index('ray-optics')
    models_dir = Path(*pth.parts[:root_pos+1]) / 'models'
    filepath = models_dir / 'Sasian Triplet.roa'

    def cff(**kwargs):
        return create_from_file(filepath, **kwargs)

    cmds.append(('Sasian Triplet',
                 (dgm.register_add_replace_element, (),
                  {'filename': filepath,
                   'node_init': create_thinlens,
                   'factory': cff,
                   'interact_mode': 'transmit'})))

    return cmds


class Diagram():
    """ class for paraxial ray rendering/editing """

    def __init__(self, opt_model, dgm_type, seq_start=1,
                 do_barrel_constraint=False, barrel_constraint=1.0,
                 label='paraxial', bend_or_gap='bend', is_dark=True):
        self.label = label
        self.opt_model = opt_model

        self.dgm_rgb = light_or_dark(is_dark=is_dark)

        self.dgm_type = dgm_type
        self.setup_dgm_type(dgm_type)

        self.do_barrel_constraint = do_barrel_constraint
        self.barrel_constraint_radius = barrel_constraint

        self.bend_or_gap = bend_or_gap

    def setup_dgm_type(self, dgm_type):
        parax_model = self.opt_model.parax_model
        if dgm_type == 'ht':
            self.type_sel = ht
            self._apply_data = parax_model.apply_ht_dgm_data
        elif dgm_type == 'slp':
            self.type_sel = slp
            self._apply_data = parax_model.apply_slope_dgm_data

    def get_label(self):
        return self.label

    def sync_light_or_dark(self, is_dark):
        self.dgm_rgb = light_or_dark(is_dark)

    def update_data(self, fig):
        parax_model = self.opt_model.parax_model

        self.shape = self.render_shape()
        self.shape_bbox = bbox_from_poly(self.shape)
        if fig.build == 'update':
            # just a change in node positions - handled above
            # self.shape = self.render_shape()
            pass
        else:
            # number of nodes have changed, rebuild everything
            if fig.build == 'full_rebuild':
                # change from another non-parax_model source,
                #  rebuild parax_model
                parax_model.build_lens()

            self.node_list = []
            for i in range(len(parax_model.sys)):
                self.node_list.append(DiagramNode(self, i))

            self.edge_list = []
            for i in range(len(parax_model.sys)-1):
                self.edge_list.append(DiagramEdge(self, i))

            self.object_shift = ConjugateLine(self, 'object_image')
            if self.opt_model.seq_model.stop_surface is None:
                # stop shift conjugate line is only editable for floating stop
                self.stop_shift = ConjugateLine(self, 'stop')

            if self.do_barrel_constraint:
                self.barrel_constraint = BarrelConstraint(self)

        self.node_bbox = fig.update_patches(self.node_list)
        self.edge_bbox = fig.update_patches(self.edge_list)

        self.object_shift_bbox = fig.update_patches([self.object_shift])
        if self.opt_model.seq_model.stop_surface is None:
            self.stop_shift_bbox = fig.update_patches([self.stop_shift])

        if self.do_barrel_constraint:
            self.barrel_bbox = fig.update_patches([self.barrel_constraint])

        return self.shape_bbox

    def apply_data(self, node, vertex):
        self._apply_data(node, vertex)
        self.opt_model.parax_model.paraxial_lens_to_seq_model()

    def assign_object_to_node(self, node, factory, **kwargs):
        parax_model = self.opt_model.parax_model
        inputs = parax_model.assign_object_to_node(node, factory, **kwargs)
        parax_model.paraxial_lens_to_seq_model()
        return inputs

    def register_commands(self, *args, **inputs):
        fig = inputs.pop('figure')
        self.command_inputs = dict(inputs)

        def do_command_action(event, target, event_key):
            nonlocal fig
            if target is not None:
                shape, handle = target.artist.shape
                try:
                    # print(type(shape).__name__, shape.node, handle, event_key)
                    handle_action_obj = shape.actions[handle]
                    if isinstance(handle_action_obj, dict):
                        handle_action_obj[event_key](fig, event)
                    else:
                        handle_action_obj.actions[event_key](fig, event)
                except KeyError:
                    pass
        fig.do_action = do_command_action

    def register_add_replace_element(self, *args, **inputs):
        fig = inputs.pop('figure')
        self.command_inputs = dict(inputs)
        action_obj = AddReplaceElementAction(self, **inputs)

        def do_command_action(event, target, event_key):
            nonlocal fig
            shape, handle = target.artist.shape
            if isinstance(shape, DiagramNode) or \
               isinstance(shape, DiagramEdge):
                try:
                    action_obj.actions[event_key](fig, event, shape)
                except KeyError:
                    pass
        fig.do_action = do_command_action

    def render_shape(self):
        """ render the diagram into a shape list """
        parax_model = self.opt_model.parax_model
        shape = []
        for x, y in zip(parax_model.pr, parax_model.ax):
            shape.append([x[self.type_sel], y[self.type_sel]])
        shape = np.array(shape)
        return shape

    def update_diagram_from_shape(self, shape):
        ''' use the shape list to update the paraxial model '''
        parax_model = self.opt_model.parax_model
        for x, y, shp in zip(parax_model.pr, parax_model.ax, shape):
            x[self.type_sel] = shp[0]
            y[self.type_sel] = shp[1]

    def fit_axis_limits(self):
        ''' define diagram axis limits as the extent of the shape polygon '''
        x_min, x_max = fit_data_range([x[0] for x in self.shape])
        y_min, y_max = fit_data_range([x[1] for x in self.shape])
        return np.array([[x_min, y_min], [x_max, y_max]])


def compute_slide_line(shape, node, imode):
    """ compute a constraint line to keep the overall length of the airspaces
        surrounding `node` constant
    """
    def distance_sqr(pt):
        return pt[0]**2 + pt[1]**2
    num_nodes = len(shape)
    if node > 0 and node < num_nodes-1:
        pt0 = shape[node-1]
        pt1 = shape[node]
        pt2 = shape[node+1]
        if imode == 'transmit':
            # if transmitting, constrain movement of node to a line parallel
            #  to the line between the previous and next nodes.
            origin2line = misc_math.perpendicular_from_origin(pt0, pt2)
            pt2line = misc_math.perpendicular_to_line(pt1, pt0, pt2)
            scale_factor = (origin2line - pt2line)/origin2line
            new_pt0 = scale_factor*pt0
            new_pt2 = scale_factor*pt2
            return new_pt0, new_pt2
        elif imode == 'reflect':
            # if reflecting, constrain movement to a radial line from the
            #  origin through the selected node
            origin = np.array([0., 0.])
            dist_pt0 = distance_sqr(pt0)
            dist_pt1 = distance_sqr(pt1)
            dist_pt2 = distance_sqr(pt2)
            # figure out how long a line to draw
            dist = dist_pt0 if dist_pt0 > dist_pt1 else dist_pt1
            dist = dist_pt2 if dist_pt2 > dist else dist
            scale_factor = math.sqrt(dist/dist_pt1)
            new_pt1 = scale_factor*pt1
            return origin, new_pt1
    return None


def constrain_to_line_action(pt0, pt2):

    def constrain_to_line(input_pt):
        """ constrain the input point to the line pt0 to pt2 """
        output_pt = misc_math.projected_point_on_line(input_pt, pt0, pt2)
        return output_pt

    return constrain_to_line


class DiagramNode():
    def __init__(self, diagram, idx):
        self.diagram = diagram
        self.node = idx
        self.select_pt = None
        self.move_direction = None
        self.handles = {}
        self.actions = self.handle_actions()

    def update_shape(self, view):
        node_color = rgb2mpl([138, 43, 226])  # blueviolet
        sys = self.diagram.opt_model.parax_model.sys
        shape = self.diagram.shape
        dgm_rgb = self.diagram.dgm_rgb
        self.handles['shape'] = (shape[self.node], 'vertex',
                                 {'linestyle': '',
                                  'linewidth': 3,
                                  'marker': 's',
                                  'picker': 6,
                                  'color': dgm_rgb['node'],
                                  # 'hilite': 'red',
                                  'zorder': 3.})
        # define the "constant spacing" or "slide" constraint
        if view.enable_slide:
            slide_pts = compute_slide_line(shape, self.node,
                                           sys[self.node][rmd])
            if slide_pts is not None:
                seg = [*slide_pts]
                f_color = rgb2mpl([138, 43, 226, 127])  # blueviolet
                h_color = rgb2mpl([138, 43, 226, 255])  # blueviolet

                self.handles['slide'] = (seg, 'polyline',
                                         {'linestyle': ':',
                                          'linewidth': 3,
                                          'picker': 6,
                                          'color': dgm_rgb['slide'],
                                          # 'hilite': h_color,
                                          'zorder': 2.5})
        gui_handles = {}
        for key, graphics_handle in self.handles.items():
            poly_data, poly_type, kwargs = graphics_handle
            poly = np.array(poly_data)
            if poly_type == 'vertex':
                p = view.create_vertex(poly, **kwargs)
            elif poly_type == 'polyline':
                p = view.create_polyline(poly, **kwargs)
            elif poly_type == 'polygon':
                p = view.create_polygon(poly, self.render_color(),
                                        **kwargs)
            else:
                break
            if len(poly.shape) > 1:
                bbox = bbox_from_poly(poly)
            else:
                x = poly[0]
                y = poly[1]
                bbox = np.array([[x, y], [x, y]])
            gui_handles[key] = GUIHandle(p, bbox)
        return gui_handles

    def render_color(self):
        e = self.diagram.opt_model.ele_model.elements[self.node]
        return e.render_color

    def get_label(self):
        return 'node' + str(self.node)

    def handle_actions(self):
        actions = {}
        actions['shape'] = EditNodeAction(self)
        slide_filter = None
        sys = self.diagram.opt_model.parax_model.sys
        slide_pts = compute_slide_line(self.diagram.shape, self.node,
                                       sys[self.node][rmd])
        if slide_pts is not None:
            slide_filter = constrain_to_line_action(*slide_pts)
        actions['slide'] = EditNodeAction(self, filter=slide_filter)
        return actions


class DiagramEdge():
    def __init__(self, diagram, idx):
        self.diagram = diagram
        self.node = idx
        self.select_pt = None
        self.move_direction = None
        self.handles = {}
        self.actions = self.handle_actions()

    def update_shape(self, view):
        shape = self.diagram.shape
        dgm_rgb = self.diagram.dgm_rgb
        edge_poly = shape[self.node:self.node+2]
        self.handles['shape'] = (edge_poly, 'polyline',
                                 {'picker': 6,
                                  'linewidth': 3,
                                  'color': dgm_rgb['edge'],
                                  # 'hilite': 'red',
                                  'zorder': 2.})
        area_poly = [[0, 0]]
        area_poly.extend(edge_poly)
        fill_color = self.render_color()
        self.handles['area'] = (area_poly, 'polygon',
                                {'fill_color': fill_color,
                                 'zorder': 1.})

        gui_handles = {}
        for key, graphics_handle in self.handles.items():
            poly_data, poly_type, kwargs = graphics_handle
            poly = np.array(poly_data)
            if poly_type == 'polygon':
                p = view.create_polygon(poly, self.render_color(), **kwargs)
            elif poly_type == 'polyline':
                p = view.create_polyline(poly, **kwargs)
            else:
                break
            gui_handles[key] = GUIHandle(p, bbox_from_poly(poly))
        return gui_handles

    def render_color(self):
        gap = self.diagram.opt_model.seq_model.gaps[self.node]
        e = self.diagram.opt_model.ele_model.gap_dict.get(gap)
        if hasattr(e, 'gap'):
            if isinstance(e, AirGap):
                # set alpha to 25% -> #40
                bkgrnd_rbga = self.diagram.dgm_rgb['background1'] + '40'
                return bkgrnd_rbga
            else:
                return e.render_color
        else:
            # single surface element, like mirror or thinlens, use airgap
            # set alpha to 25% -> #40
            bkgrnd_rbga = self.diagram.dgm_rgb['background1'] + '40'
            return bkgrnd_rbga

    def get_label(self):
        return 'edge' + str(self.node)

    def handle_actions(self):
        actions = {}
        actions['shape'] = EditLensAction(self)
        return actions


class BarrelConstraint():
    def __init__(self, diagram):
        self.diagram = diagram
        self.handles = {}
        self.actions = self.handle_actions()

    def update_shape(self, view):
        dgm_rgb = self.diagram.dgm_rgb
        barrel_radius = self.diagram.barrel_constraint_radius
        diamond = []
        diamond.append([0., barrel_radius])
        diamond.append([barrel_radius, 0.])
        diamond.append([0., -barrel_radius])
        diamond.append([-barrel_radius, 0.])
        diamond.append([0., barrel_radius])
        diamond = np.array(diamond)
        self.handles['shape'] = (diamond, 'polyline',
                                 {'color': dgm_rgb['barrel'],
                                  'linewidth': 3,
                                  'zorder': 1.})
        square = []
        square.append([ barrel_radius,  barrel_radius])
        square.append([-barrel_radius,  barrel_radius])
        square.append([-barrel_radius, -barrel_radius])
        square.append([ barrel_radius, -barrel_radius])
        square.append([ barrel_radius,  barrel_radius])
        square = np.array(square)
        self.handles['square'] = (square, 'polyline',
                                  {'color': dgm_rgb['barrel'],
                                   'linewidth': 3,
                                   'zorder': 1.})

        gui_handles = {}
        for key, graphics_handle in self.handles.items():
            poly_data, poly_type, kwargs = graphics_handle
            poly = np.array(poly_data)
            if poly_type == 'polygon':
                p = view.create_polygon(poly, self.render_color(), **kwargs)
            elif poly_type == 'polyline':
                p = view.create_polyline(poly, **kwargs)
            else:
                break
            gui_handles[key] = GUIHandle(p, bbox_from_poly(poly))
        return gui_handles

    def render_color(self):
        return 'black'

    def get_label(self):
        return 'barrel constraint'

    def handle_actions(self):
        actions = {}
        return actions


class ConjugateLine():
    def __init__(self, diagram, line_type):
        self.diagram = diagram
        self.line_type = line_type
        self.sys_orig = []
        self.shape_orig = []
        self.handles = {}
        self.actions = self.handle_actions()

    def update_shape(self, view):
        dgm_rgb = self.diagram.dgm_rgb
        shape_bbox = self.diagram.shape_bbox
        line = []
        if self.line_type == 'stop':
            ht = shape_bbox[1][1] - shape_bbox[0][1]
            line.append([0., -2*ht])
            line.append([0., 2*ht])
            color = dgm_rgb['stop']
        elif self.line_type == 'object_image':
            wid = shape_bbox[1][0] - shape_bbox[0][0]
            line.append([-2*wid, 0.])
            line.append([2*wid, 0.])
            color = dgm_rgb['object_image']
        self.handles['shape'] = (line, 'polyline',
                                 {'color': dgm_rgb['foreground'],
                                  'hilite': color,
                                  'zorder': 1.})

        if len(self.shape_orig) > 0:
            conj_line = []
            if self.line_type == 'stop':
                lwr, upr = view.ax.get_ybound()
                ht = upr - lwr
                conj_line.append([self.k*ht, -ht])
                conj_line.append([-self.k*ht, ht])
            elif self.line_type == 'object_image':
                lwr, upr = view.ax.get_xbound()
                wid = upr - lwr
                conj_line.append([-wid, self.k*wid])
                conj_line.append([wid, -self.k*wid])
            self.handles['conj_line'] = (conj_line, 'polyline',
                                         {'color': dgm_rgb['conj_line'],
                                          'zorder': 1.})
            self.handles['shift'] = (self.shape_orig, 'polyline',
                                     {'color': dgm_rgb['shift'],
                                      'zorder': 1.})

        gui_handles = {}
        for key, graphics_handle in self.handles.items():
            poly_data, poly_type, kwargs = graphics_handle
            poly = np.array(poly_data)
            if poly_type == 'polygon':
                p = view.create_polygon(poly, self.render_color(), **kwargs)
            elif poly_type == 'polyline':
                p = view.create_polyline(poly, **kwargs)
            else:
                break
            gui_handles[key] = GUIHandle(p, bbox_from_poly(poly))
        return gui_handles

    def render_color(self):
        return 'black'

    def get_label(self):
        if self.line_type == 'stop':
            return 'stop shift line'
        elif self.line_type == 'object_image':
            return 'object shift line'
        else:
            return ''

    def edit_conjugate_line_actions(self):
        dgm = self.diagram
        pm = dgm.opt_model.parax_model

        def calculate_slope(x, y):
            ''' x=ybar, y=y  '''
            if self.line_type == 'stop':
                k = -x/y
                return k, np.array([[1, 0], [k, 1]])
            elif self.line_type == 'object_image':
                k = -y/x
                return k, np.array([[1, k], [0, 1]])
            else:
                return 0, np.array([[1, 0], [0, 1]])

        def apply_data(event_data):
            self.k, mat = calculate_slope(event_data[0], event_data[1])
            # apply the shear transformation to the original shape
            dgm.shape = self.shape_orig.dot(mat)

            dgm.update_diagram_from_shape(dgm.shape)

            # update slope values
            for i in range(1, len(dgm.shape)):
                pm.pr[i-1][slp] = ((pm.pr[i][ht] - pm.pr[i-1][ht]) /
                                   self.sys_orig[i-1][tau])
                pm.ax[i-1][slp] = ((pm.ax[i][ht] - pm.ax[i-1][ht]) /
                                   self.sys_orig[i-1][tau])
            pm.pr[-1][slp] = pm.pr[-2][slp]
            pm.ax[-1][slp] = pm.ax[-2][slp]

            if self.line_type == 'object_image':
                # update object distance and object y and ybar
                pm.sys[0][tau] = pm.ax[1][ht]/pm.ax[0][slp]
                pm.ax[0][ht] = 0
                pm.pr[0][ht] = pm.pr[1][ht] - pm.sys[0][tau]*pm.pr[0][slp]

                # update image distance and image y and ybar
                pm.sys[-2][tau] = -pm.ax[-2][ht]/pm.ax[-2][slp]
                pm.ax[-1][ht] = 0
                pm.pr[-1][ht] = pm.pr[-2][ht] + pm.sys[-2][tau]*pm.pr[-2][slp]

            pm.paraxial_lens_to_seq_model()

        def on_select(fig, event):
            self.sys_orig = deepcopy(pm.sys)
            self.shape_orig = deepcopy(dgm.shape)

        def on_edit(fig, event):
            if event.xdata is not None and event.ydata is not None:
                event_data = np.array([event.xdata, event.ydata])
                apply_data(event_data)
                fig.build = 'update'
                fig.refresh_gui()

        def on_release(fig, event):
            event_data = np.array([event.xdata, event.ydata])
            apply_data(event_data)
            self.sys_orig = []
            self.shape_orig = []
            fig.refresh_gui()

        actions = {}
        actions['press'] = on_select
        actions['drag'] = on_edit
        actions['release'] = on_release
        return actions

    def handle_actions(self):
        actions = {}
        actions['shape'] = self.edit_conjugate_line_actions()
        return actions


class EditNodeAction():
    """ Action to move a diagram node, using an input pt """

    def __init__(self, dgm_node, filter=None):
        diagram = dgm_node.diagram
        self.cur_node = None
        self.pt0 = None
        self.pt2 = None
        self.filter = filter

        def point_on_line(pt1, pt2, t):
            d = pt2 - pt1
            return pt1 + t*d

        def constrain_to_wedge(input_pt):
            """ keep the input point inside the wedge of adjacent points """

            if self.pt0 is not None:
                x_prod0 = input_pt[0]*self.pt0[1] - self.pt0[0]*input_pt[1]
                if x_prod0 < 0:
                    # pin to boundary
                    output_pt = projected_point_on_radial_line(input_pt,
                                                               self.pt0)
                    return output_pt

            if self.pt2 is not None:
                x_prod2 = input_pt[0]*self.pt2[1] - self.pt2[0]*input_pt[1]
                if x_prod2 > 0:
                    # pin to boundary
                    output_pt = projected_point_on_radial_line(input_pt,
                                                               self.pt2)
                    return output_pt

            return input_pt

        def on_select(fig, event):
            nonlocal diagram
            # we don't allow points to move onto their adjacent neighbors. Use
            #  a buffer amount when constraining to the wedge
            buffer_fraction = 0.0025
            self.cur_node = dgm_node.node
            pt1 = diagram.shape[self.cur_node]
            if self.cur_node == 0:
                pt2 = diagram.shape[self.cur_node+1]
                self.pt2 = point_on_line(pt1, pt2, 1-buffer_fraction)
                self.pt0 = None
            elif self.cur_node == len(diagram.shape)-1:
                pt0 = diagram.shape[self.cur_node-1]
                self.pt0 = point_on_line(pt0, pt1, buffer_fraction)
                self.pt2 = None
            else:
                pt0 = diagram.shape[self.cur_node-1]
                self.pt0 = point_on_line(pt0, pt1, buffer_fraction)
                pt2 = diagram.shape[self.cur_node+1]
                self.pt2 = point_on_line(pt1, pt2, 1-buffer_fraction)

        def on_edit(fig, event):
            if event.xdata is not None and event.ydata is not None:
                event_data = np.array([event.xdata, event.ydata])
                if self.filter:
                    event_data = self.filter(event_data)
                event_data = constrain_to_wedge(event_data)
                diagram.apply_data(self.cur_node, event_data)
                fig.build = 'update'
                fig.refresh_gui()

        def on_release(fig, event):
            if event.xdata is not None and event.ydata is not None:
                event_data = np.array([event.xdata, event.ydata])
                if self.filter:
                    event_data = self.filter(event_data)
                event_data = constrain_to_wedge(event_data)
                diagram.apply_data(self.cur_node, event_data)
                fig.build = 'update'
                fig.refresh_gui()
                self.cur_node = None

        self.actions = {}
        self.actions['drag'] = on_edit
        self.actions['press'] = on_select
        self.actions['release'] = on_release


class EditLensAction():
    """ Action for diagram edge, using an input pt

    This is a simple wrapper class to choose the correct action, i.e. bending
    or thickness change, depending on the UI setting.
    """

    def __init__(self, dgm_edge):
        diagram = dgm_edge.diagram
        actions = {}
        actions['gap'] = EditThicknessAction(dgm_edge)
        actions['bend'] = EditBendingAction(dgm_edge)

        def create_dispatch_action(event_key):
            def dispatch_action(fig, event):
                actions[diagram.bend_or_gap].actions[event_key](fig, event)
            return dispatch_action

        self.actions = {}
        self.actions['press'] = create_dispatch_action('press')
        self.actions['drag'] = create_dispatch_action('drag')
        self.actions['release'] = create_dispatch_action('release')


class EditThicknessAction():
    """ Action to move a diagram edge, using an input pt

    The movement is constrained to be parallel to the original edge. By doing
    this the power and bending of the element remains constant, while the
    element thickness changes. Movement of the edge is limited to keep
    the thickness greater than zero and not to interfere with adjacent spaces.
    """

    def __init__(self, dgm_edge):
        diagram = dgm_edge.diagram
        self.node = None
        self.bundle = None

        def on_select(fig, event):
            nonlocal diagram
            shape = diagram.shape
            self.node = node = dgm_edge.node
            if node > 0 and node < (len(shape)-2):
                # get the virtual vertex of the combined element surfaces
                vertex = np.array(get_intersect(shape[node-1], shape[node],
                                                shape[node+1], shape[node+2]))
                # get direction cosines for the edge
                edge = normalize(shape[node+1] - shape[node])
                # construct perpendicular to the edge. use this to define a
                # range for allowed inputs
                perp_edge = np.array([edge[1], -edge[0]])
                self.bundle = vertex, edge, perp_edge
                # measure distances along the perpendicular thru the vertex
                #  and project the 2 outer nodes and the first vertex of the
                #  edge
                pp0 = np.dot(shape[node-1]-vertex, perp_edge)
                pp1 = np.dot(shape[node]-vertex, perp_edge)
                pp3 = np.dot(shape[node+2]-vertex, perp_edge)
                # use the first edge vertex to calculate an allowed input range
                if pp1 > 0:
                    self.pp_lim = 0, min(i for i in (pp0, pp3) if i > 0)
                else:
                    self.pp_lim = max(i for i in (pp0, pp3) if i < 0), 0

        def on_edit(fig, event):
            buffer = 0.0025
            nonlocal diagram
            shape = diagram.shape
            node = self.node
            if node > 0 and node < (len(shape)-2):
                if event.xdata is not None and event.ydata is not None:
                    inpt = np.array([event.xdata, event.ydata])
                    vertex, edge, perp_edge = self.bundle
                    # project input pt onto perp_edge
                    ppi = np.dot(inpt - vertex, perp_edge)
                    # constrain the input point within the range, if needed
                    if ppi < self.pp_lim[0]:
                        inpt = vertex + (1+buffer)*self.pp_lim[0]*perp_edge
                    elif ppi > self.pp_lim[1]:
                        inpt = vertex + (1-buffer)*self.pp_lim[1]*perp_edge

                    # compute new edge vertices from intersection of adjacent
                    #  edges and line shifted parallel to the initial edge
                    edge_pt = inpt + edge
                    pt1 = np.array(get_intersect(shape[node-1], vertex,
                                                 inpt, edge_pt))
                    diagram.apply_data(self.node, pt1)
                    pt2 = np.array(get_intersect(vertex, shape[node+2],
                                                 inpt, edge_pt))
                    diagram.apply_data(self.node+1, pt2)
                    fig.build = 'update'
                    fig.refresh_gui()

        def on_release(fig, event):
            if event.xdata is not None and event.ydata is not None:
                event_data = np.array([event.xdata, event.ydata])
                fig.build = 'update'
                fig.refresh_gui()
                self.node = None
                self.bundle = None

        self.actions = {}
        self.actions['drag'] = on_edit
        self.actions['press'] = on_select
        self.actions['release'] = on_release


class EditBendingAction():
    """ Action to bend the lens element for diagram edge, using an input pt.

    The movement is constrained to be along the object ray for the lens if the
    input point is closer to the leading node of the edge. Otherwise the
    movement is constrained to be along the image ray. The unconstrained point
    is solved to keep the element thickness constant and maintain the
    object-image properties of the lens.
    """

    def __init__(self, dgm_edge):
        diagram = dgm_edge.diagram
        pm = diagram.opt_model.parax_model

        self.node = None
        self.bundle = None
        self.filter = None

        def cross_prod(pt1, pt2):
            return pt1[1]*pt2[0] - pt2[1]*pt1[0]

        def calc_coef_fct(vertex, iNode, dir_inpt, oNode, dir_out):
            nonlocal pm
            tau_factor = pm.sys[self.node][tau]*pm.opt_inv
            constrain_to_line = constrain_to_line_action(vertex,
                                                         vertex+dir_inpt)

            def calc_t(inpt):
                pt = constrain_to_line(inpt)
                if iNode < oNode:
                    arg1 = pt, vertex
                    arg2 = pt, dir_out
                else:
                    arg1 = vertex, pt
                    arg2 = dir_out, pt
                t = (tau_factor - cross_prod(*arg1))/(cross_prod(*arg2))
                return (iNode, pt), (oNode, vertex + t*dir_out)
            return calc_t

        def on_select(fig, event):
            nonlocal diagram
            if event.xdata is None or event.ydata is None:
                return
            shape = diagram.shape
            self.node = node = dgm_edge.node
            if node > 0 and node < (len(shape)-2):
                inpt = np.array([event.xdata, event.ydata])
                # get the virtual vertex of the combined element surfaces
                vertex = np.array(get_intersect(shape[node-1], shape[node],
                                                shape[node+1], shape[node+2]))
                edge_dir_01 = normalize(shape[node] - shape[node-1])
                edge_dir_23 = normalize(shape[node+2] - shape[node+1])
                # which node is closer to the input point?
                pt1_dist = distance_sqr_2d(shape[node], inpt)
                pt2_dist = distance_sqr_2d(shape[node+1], inpt)
                if pt1_dist < pt2_dist:
                    self.filter = calc_coef_fct(vertex, node, edge_dir_01,
                                                node+1, edge_dir_23)
                else:
                    self.filter = calc_coef_fct(vertex, node+1, edge_dir_23,
                                                node, edge_dir_01)
                # get direction cosines for the edge
                edge = normalize(shape[node+1] - shape[node])
                # construct perpendicular to the edge. use this to define a
                # range for allowed inputs
                perp_edge = np.array([edge[1], -edge[0]])
                self.bundle = (vertex, edge, perp_edge, edge_dir_01,
                               edge_dir_23)

        def on_edit(fig, event):
            nonlocal diagram
            shape = diagram.shape
            node = self.node
            if node > 0 and node < (len(shape)-2):
                if event.xdata is not None and event.ydata is not None:
                    inpt = np.array([event.xdata, event.ydata])
                    inp, out = self.filter(inpt)

                    diagram.apply_data(inp[0], inp[1])
                    diagram.apply_data(out[0], out[1])
                    fig.refresh_gui()

        def on_release(fig, event):
            if event.xdata is not None and event.ydata is not None:
                event_data = np.array([event.xdata, event.ydata])
                fig.build = 'update'
                fig.refresh_gui()
                self.node = None
                self.bundle = None
                self.filter = None

        self.actions = {}
        self.actions['drag'] = on_edit
        self.actions['press'] = on_select
        self.actions['release'] = on_release


class AddReplaceElementAction():
    ''' insert or replace a node with a chunk from a factory fct

    The do_command_action fct registered for this operation passes the shape
    being operated upon; these can be:

        - DiagramEdge: insert/add the chunk returned by the factory fct
        - DiagramNode: replace the selected node with the factory fct output

    Inserting is done by splitting the corresponding gap in two. A new gap
    and an AirGap element are tacked on to the chunk returned from the factory
    fct.
    Replacing is done when a DiagramNode is selected. The gaps surrounding the
    node are retained, and modified as needed to accomodate the chunk.
    '''

    def __init__(self, diagram, **kwargs):
        seq_model = diagram.opt_model.seq_model
        parax_model = diagram.opt_model.parax_model
        self.cur_node = None

        def on_press_add_point(fig, event, shape):
            # if we don't have factory functions, skip the command
            if isinstance(shape, DiagramEdge):
                if 'node_init' in diagram.command_inputs and \
                   'factory' in diagram.command_inputs:

                    self.cur_node = shape.node
                    event_data = np.array([event.xdata, event.ydata])
                    interact = diagram.command_inputs['interact_mode']
                    parax_model.add_node(self.cur_node, event_data,
                                         diagram.type_sel, interact)
                    self.cur_node += 1
                    # create a node for editing during the drag action
                    #  'node_init' will currently be a thinlens or a mirror
                    node_init = diagram.command_inputs['node_init']
                    self.init_inputs = diagram.assign_object_to_node(
                                            self.cur_node,
                                            node_init,
                                            insert=True)
                    fig.build = 'rebuild'
                    fig.refresh_gui()
            elif isinstance(shape, DiagramNode):
                if 'factory' in diagram.command_inputs:
                    # replacing a node with a chunk only requires recording
                    # what chunk corresponds to the current node. There is
                    # no drag action
                    self.cur_node = node = shape.node
                    self.init_inputs = parax_model.get_object_for_node(node)

        def on_drag_add_point(fig, event, shape):
            if self.cur_node is not None and isinstance(shape, DiagramEdge):
                event_data = np.array([event.xdata, event.ydata])
                diagram.apply_data(self.cur_node, event_data)
                fig.build = 'update'
                fig.refresh_gui()

        def on_release_add_point(fig, event, shape):
            if self.cur_node is not None:
                factory = diagram.command_inputs['factory']
                # if factory and node_init fcts are the same, we're done;
                # always call factory fct for a node
                if factory != diagram.command_inputs['node_init'] or \
                              isinstance(shape, DiagramNode):
                    prev_ifc = seq_model.ifcs[self.cur_node]
                    inputs = diagram.assign_object_to_node(self.cur_node,
                                                           factory)
                    idx = seq_model.ifcs.index(prev_ifc)
                    n_after = parax_model.sys[idx-1][indx]
                    thi = n_after*parax_model.sys[idx-1][tau]
                    seq_model.gaps[idx-1].thi = thi
                    # remove the edit scaffolding or previous node from model
                    args, kwargs = self.init_inputs
                    diagram.opt_model.remove_ifc_gp_ele(*args, **kwargs)
                fig.build = 'rebuild'
                fig.refresh_gui()
            self.cur_node = None

        self.actions = {}
        self.actions['press'] = on_press_add_point
        self.actions['drag'] = on_drag_add_point
        self.actions['release'] = on_release_add_point


class AddElementAction():
    ''' Deprecated '''

    def __init__(self, dgm_edge, **kwargs):
        diagram = dgm_edge.diagram
        seq_model = diagram.opt_model.seq_model
        parax_model = diagram.opt_model.parax_model
        self.cur_node = None

        def on_press_add_point(fig, event):
            # if we don't have factory functions, skip the command
            if 'node_init' in diagram.command_inputs and \
               'factory' in diagram.command_inputs:

                self.cur_node = dgm_edge.node
                event_data = np.array([event.xdata, event.ydata])
                interact = diagram.command_inputs['interact_mode']
                parax_model.add_node(self.cur_node, event_data,
                                     diagram.type_sel, interact)
                self.cur_node += 1
                node_init = diagram.command_inputs['node_init']
                self.init_inputs = diagram.assign_object_to_node(self.cur_node,
                                                                 node_init)
                fig.build = 'rebuild'
                fig.refresh_gui()

        def on_drag_add_point(fig, event):
            if self.cur_node is not None:
                event_data = np.array([event.xdata, event.ydata])
                diagram.apply_data(self.cur_node, event_data)
                fig.build = 'update'
                fig.refresh_gui()

        def on_release_add_point(fig, event):
            if self.cur_node is not None:
                factory = diagram.command_inputs['factory']
                if factory != diagram.command_inputs['node_init']:
                    prev_ifc = seq_model.ifcs[self.cur_node]
                    inputs = diagram.assign_object_to_node(self.cur_node,
                                                           factory)
                    idx = seq_model.ifcs.index(prev_ifc)
                    n_after = parax_model.sys[idx-1][indx]
                    thi = n_after*parax_model.sys[idx-1][tau]
                    seq_model.gaps[idx-1].thi = thi
                    args, kwargs = self.init_inputs
                    diagram.opt_model.remove_ifc_gp_ele(*args, **kwargs)
                fig.build = 'rebuild'
                fig.refresh_gui()
            self.cur_node = None

        self.actions = {}
        self.actions['press'] = on_press_add_point
        self.actions['drag'] = on_drag_add_point
        self.actions['release'] = on_release_add_point



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Support creation of an iPython console, with rayoptics environment

.. Created on Wed Nov 21 21:48:02 2018

.. codeauthor: Michael J. Hayford
"""

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager
from IPython.lib import guisupport

from rayoptics.gui.appmanager import ModelInfo


def create_ipython_console(app, title, view_width, view_ht):
    """ create a iPython console with a rayoptics environment """
    opt_model = app.app_manager.model
    if opt_model:
        ro_env = {
                'app': app,
                'opm': opt_model,
                'sm': opt_model.seq_model,
                'osp': opt_model.optical_spec,
                'pm': opt_model.parax_model
                }
    else:
        ro_env = {
            'app': app,
            'opm': opt_model
            }

    ro_setup = 'from rayoptics.environment import *'

    # construct the top level widget
    ipy_console = ConsoleWidget()

    # load the environment
    ipy_console.execute_command(ro_setup)
    ipy_console.push_vars(ro_env)

    mi = ModelInfo(opt_model)
    sub_window = app.add_subwindow(ipy_console, mi)
    sub_window.setWindowTitle(title)
    orig_x, orig_y = app.initial_window_offset()
    sub_window.setGeometry(orig_x, orig_y, view_width, view_ht)

    sub_window.show()


class ConsoleWidget(RichJupyterWidget):

    def __init__(self, customBanner=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if customBanner is not None:
            self.banner = customBanner

        self.font_size = 6
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel(show_banner=False)
        kernel_manager.kernel.gui = 'qt'
        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            guisupport.get_app_qt().exit()

        self.exit_requested.connect(stop)

    def push_vars(self, variableDict):
        """
        Given a dictionary containing name / value pairs, push those variables
        to the Jupyter console widget
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clear(self):
        """
        Clears the terminal
        """
        self._control.clear()

        # self.kernel_manager

    def print_text(self, text):
        """
        Prints some plain text to the console
        """
        self._append_plain_text(text)

    def execute_command(self, command):
        """
        Execute a command in the frame of the console widget
        """
        self._execute(command, False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""

.. Created on Fri Apr  13 10:43:19 2018

.. codeauthor: Michael J. Hayford
"""

from enum import Enum, auto

from scipy.interpolate import interp1d

import numpy as np

from rayoptics.mpl.styledfigure import StyledFigure

from rayoptics.optical.raytrace import wave_abr
import rayoptics.optical.model_constants as mc


class Fit(Enum):
    All = auto()
    All_Same = auto()
    User_Scale = auto()


def clip_to_range(rgb_list, lower, upper):
    rgbc_list = []
    for rgb in rgb_list:
        rgbc = []
        for i, clr in enumerate(rgb):
            rgbc.append(clr)
            if clr < lower:
                rgbc[i] = lower
            if upper < clr:
                rgbc[i] = upper
        rgbc_list.append(rgbc)
    return rgbc


class AxisArrayFigure(StyledFigure):

    def __init__(self, opt_model,
                 num_rays=21,
                 scale_type=Fit.All,
                 user_scale_value=0.1,
                 num_rows=1, num_cols=1,
                 eval_fct=None, **kwargs):
        self.opt_model = opt_model
        self.num_rays = num_rays
        self.user_scale_value = user_scale_value
        self.scale_type = scale_type

        super().__init__(**kwargs)

        self.eval_fct = eval_fct
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.update_data()

    def init_axis(self, ax):
        ax.grid(True)
        ax.set_xlim(-1., 1.)
        ax.axvline(0, c=self._rgb['foreground'], lw=1)
        ax.axhline(0, c=self._rgb['foreground'], lw=1)

    def construct_plot_array(self, m, n):
        arr = []
        k = 1
        for i in reversed(range(m)):
            row = []
            for j in reversed(range(n)):
                ax = self.add_subplot(m, n, k)
                self.init_axis(ax)
#                title = "["+str(i)+"],["+str(j)+"]"
#                print("title, id(ax):", title, id(ax))
#                ax.set_title(title)
                row.append(ax)
                k += 1
            arr.append(row)
        return arr

    def wvl_to_sys_units(self, wvl):
        return self.opt_model.nm_to_sys_units(wvl)

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        pass

    def plot(self):
        pass


class RayFanFigure(AxisArrayFigure):

    def __init__(self, opt_model, data_type, override_style=True,
                 **kwargs):
        self.max_value_all = 0.0
        self.override_style = override_style
        seq_model = opt_model.seq_model
        osp = opt_model.optical_spec

        def ray_abr(p, xy, ray_pkg, fld, wvl, foc):
            image_pt = fld.ref_sphere[0][0]
            if ray_pkg[mc.ray] is not None:
                ray = ray_pkg[mc.ray]
                y_val = ray[-1][mc.p][xy] - image_pt[xy]
                return y_val
            else:
                return None

        def opd(p, xy, ray_pkg, fld, wvl, foc):
            if ray_pkg[mc.ray] is not None:
                opd = wave_abr(fld, wvl, foc, ray_pkg)
                return opd[0]/self.wvl_to_sys_units(wvl)
            else:
                return None

        def eval_abr_fan(i, j):
            fld, wvl, foc = osp.lookup_fld_wvl_focus(i)
            return seq_model.trace_fan(ray_abr, i, j, num_rays=self.num_rays)

        def eval_opd_fan(i, j):
            fld, wvl, foc = osp.lookup_fld_wvl_focus(i)
            return seq_model.trace_fan(opd, i, j, num_rays=self.num_rays)

        if data_type == 'Ray':
            eval_fan = eval_abr_fan
        elif data_type == 'OPD':
            eval_fan = eval_opd_fan
        num_flds = len(osp.field_of_view.fields)
        super().__init__(opt_model, eval_fct=eval_fan,
                         num_rows=num_flds, num_cols=2, **kwargs)

    def update_data(self):
        self.axis_data_array = []
        for i in reversed(range(self.num_rows)):
            row = []
            for j in reversed(range(self.num_cols)):
                x_smooth = []
                y_smooth = []
                x_data, y_data, max_value, rc = self.eval_fct(i, j)
#                x_data, y_data, max_value, rc = self.eval_axis_data(i, j)
#                rc = clip_to_range(rc, 0.0, 1.0)
                for k in range(len(x_data)):
                    interpolator = interp1d(x_data[k], y_data[k],
                                            kind='cubic', assume_sorted=True)
                    x_samp = np.linspace(x_data[k].min(), x_data[k].max(), 100)
                    y_fit = interpolator(x_samp)
                    x_smooth.append(x_samp)
                    y_smooth.append(y_fit)
                row.append((x_smooth, y_smooth, max_value, rc))
            self.axis_data_array.append(row)
        return self

    def plot(self):
        if hasattr(self, 'ax_arr'):
            self.clf()

        m = self.num_rows - 1
        n = self.num_cols - 1
        self.ax_arr = self.construct_plot_array(self.num_rows, self.num_cols)

        self.max_value_all = 0.0
        for i in reversed(range(self.num_rows)):
            for j in reversed(range(self.num_cols)):
                x_data, y_data, max_value, rc = self.axis_data_array[m-i][n-j]
                ax = self.ax_arr[m-i][n-j]
                for k in range(len(x_data)):
                    if self.override_style:
                        ax.plot(x_data[k], y_data[k], c=rc[k])
                    else:
                        ax.plot(x_data[k], y_data[k])

                if max_value > self.max_value_all:
                    self.max_value_all = max_value

        if self.scale_type == Fit.All:
            pass
#            print("Fit.All", self.max_value_all)
#            [[ax.set_ylim(-mv, mv) for ax in r] for r in self.ax_arr]
        if self.scale_type == Fit.All_Same:
            mv = self.max_value_all
#            print("Fit.All_Same", mv)
            [[ax.set_ylim(-mv, mv) for ax in r] for r in self.ax_arr]
        if self.scale_type == Fit.User_Scale and self.user_scale_value is not None:
            us = self.user_scale_value
#            print("User_Scale", us)
            [[ax.set_ylim(-us, us) for ax in r] for r in self.ax_arr]

        self.canvas.draw()

        return self


class SpotDiagramFigure(AxisArrayFigure):

    def __init__(self, opt_model, **kwargs):
        self.max_value_all = 0.0
        seq_model = opt_model.seq_model
        osp = opt_model.optical_spec

        def spot(p, wi, ray_pkg, fld, wvl, foc):
            image_pt = fld.ref_sphere[0][0]
#            image_pt = fld.chief_ray.ray[-1][0]
            if ray_pkg is not None:
                ray = ray_pkg[mc.ray]
                x = ray[-1][mc.p][0] - image_pt[0]
                y = ray[-1][mc.p][1] - image_pt[1]
                return np.array([x, y])
            else:
                return None

        def eval_grid(i, j):
            fld, wvl, foc = osp.lookup_fld_wvl_focus(i, wl=j)
            return seq_model.trace_grid(spot, i, num_rays=self.num_rays,
                                        form='list', append_if_none=False)

        num_flds = len(osp.field_of_view.fields)
        super().__init__(opt_model, eval_fct=eval_grid,
                         num_rows=num_flds, num_cols=1, **kwargs)

    def init_axis(self, ax):
        ax.grid(True)
        ax.axvline(0, c='black', lw=1)
        ax.axhline(0, c='black', lw=1)

    def update_data(self):
        self.axis_data_array = []
        for i in reversed(range(self.num_rows)):
            row = []
            for j in reversed(range(self.num_cols)):
                grids, rc = self.eval_fct(i, j)
                max_val = max([max(np.max(g), -np.min(g)) for g in grids])
                row.append((grids, max_val, rc))
            self.axis_data_array.append(row)
        return self

    def plot(self):
        if hasattr(self, 'ax_arr'):
            self.clf()

        m = self.num_rows - 1
        n = self.num_cols - 1
        self.ax_arr = self.construct_plot_array(self.num_rows, self.num_cols)

        self.max_value_all = 0.0
        for i in reversed(range(self.num_rows)):
            for j in reversed(range(self.num_cols)):
                grids, max_value, rc = self.axis_data_array[m-i][n-j]
                ax = self.ax_arr[m-i][n-j]
                for k in range(len(rc)):
                    ax.plot(np.transpose(grids[k])[0],
                            np.transpose(grids[k])[1],
                            c=rc[k], linestyle='None',
                            marker='o', markersize=2)
                    ax.set_aspect('equal')

                if max_value > self.max_value_all:
                    self.max_value_all = max_value

        if self.scale_type == Fit.All:
            pass
#            print("Fit.All", self.max_value_all)
#            [[ax.set_ylim(-mv, mv) for ax in r] for r in self.ax_arr]
        if self.scale_type == Fit.All_Same:
            mv = self.max_value_all
#            print("Fit.All_Same", mv)
            [[ax.set_xlim(-mv, mv) for ax in r] for r in self.ax_arr]
            [[ax.set_ylim(-mv, mv) for ax in r] for r in self.ax_arr]
        if self.scale_type == Fit.User_Scale and self.user_scale_value is not None:
            us = self.user_scale_value
#            print("User_Scale", us)
            [[ax.set_xlim(-us, us) for ax in r] for r in self.ax_arr]
            [[ax.set_ylim(-us, us) for ax in r] for r in self.ax_arr]

        self.tight_layout()

        self.canvas.draw()

        return self


class WavefrontFigure(AxisArrayFigure):

    def __init__(self, opt_model, **kwargs):
        self.max_value_all = 0.0
        seq_model = opt_model.seq_model
        osp = opt_model.optical_spec

        def wave(p, wi, ray_pkg, fld, wvl, foc):
            x = p[0]
            y = p[1]
            if ray_pkg is not None:
                opd_pkg = wave_abr(fld, wvl, foc, ray_pkg)
                opd = opd_pkg[0]/self.wvl_to_sys_units(wvl)
            else:
                opd = 0.0
            return np.array([x, y, opd])

        def eval_grid(i, j):
            fld, wvl, foc = osp.lookup_fld_wvl_focus(i, wl=j)
            return seq_model.trace_grid(wave, i, wl=j, num_rays=self.num_rays,
                                        form='grid', append_if_none=True)

        num_flds = len(osp.field_of_view.fields)
        num_wvls = len(osp.spectral_region.wavelengths)
        super().__init__(opt_model, eval_fct=eval_grid,
                         num_rows=num_flds, num_cols=num_wvls, **kwargs)

    def init_axis(self, ax):
#        ax.grid(True)
        ax.set_xlim(-1., 1.)
        ax.set_ylim(-1., 1.)
#        ax.axvline(0, c='black', lw=1)
#        ax.axhline(0, c='black', lw=1)

    def update_data(self):
        self.axis_data_array = []
        self.max_value_all = 0.0
        for i in reversed(range(self.num_rows)):
            row = []
            for j in reversed(range(self.num_cols)):
                grids, rc = self.eval_fct(i, j)
                g = grids[0]
                g = np.rollaxis(g, 2)
                max_value = max(np.max(g[2]), -np.min(g[2]))
                row.append((g, max_value, rc))

                if max_value > self.max_value_all:
                    self.max_value_all = max_value

            self.axis_data_array.append(row)
        return self

    def plot(self):
        if hasattr(self, 'ax_arr'):
            self.clf()

        m = self.num_rows - 1
        n = self.num_cols - 1
        self.ax_arr = self.construct_plot_array(self.num_rows, self.num_cols)

        for i in reversed(range(self.num_rows)):
            for j in reversed(range(self.num_cols)):
                grid, max_value, rc = self.axis_data_array[m-i][n-j]
                if self.scale_type == Fit.All_Same:
                    max_value = self.max_value_all
                elif (self.scale_type == Fit.User_Scale and
                      self.user_scale_value is not None):
                    max_value = self.user_scale_value
                ax = self.ax_arr[m-i][n-j]
#                hmap = ax.contourf(grid[0],
#                                   grid[1],
#                                   grid[2],
#                                   cmap="RdBu_r",
#                                   vmin=-max_value,
#                                   vmax=max_value)
                hmap = ax.imshow(grid[2],
                                 cmap="RdBu_r",
                                 vmin=-max_value,
                                 vmax=max_value,
                                 extent=[-1., 1., -1., 1.],
                                 origin='lower')
                self.colorbar(hmap, ax=ax)

                ax.set_aspect('equal')

        self.tight_layout()

        self.canvas.draw()

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Module for element modeling

.. Created on Sun Jan 28 16:27:01 2018

.. codeauthor: Michael J. Hayford
"""

import logging
from collections import namedtuple

import math
import numpy as np

import rayoptics.util.rgbtable as rgbt
import rayoptics.optical.thinlens as thinlens
from rayoptics.optical.profiles import Spherical, Conic
from rayoptics.optical.surface import Surface
from rayoptics.optical.gap import Gap
import rayoptics.gui.appcmds as cmds
from rayoptics.gui.actions import Action, AttrAction, SagAction, BendAction
from rayoptics.optical.medium import Glass, glass_decode
import rayoptics.optical.model_constants as mc
import opticalglass.glasspolygons as gp

GraphicsHandle = namedtuple('GraphicsHandle', ['polydata', 'tfrm', 'polytype'])
""" tuple grouping together graphics rendering data

    Attributes:
        polydata: poly data in local coordinates
        tfrm: global transformation for polydata
        polytype: 'polygon' (for filled) or 'polyline'
"""


def create_thinlens(power=0., indx=1.5, sd=None, **kwargs):
    tl = thinlens.ThinLens(power=power, ref_index=indx, max_ap=sd, **kwargs)
    tle = ThinElement(tl)
    return [[tl, None, None, 1, +1]], [tle]


def create_mirror(c=0.0, r=None, cc=0.0, ec=None,
                  power=None, profile=None, sd=None, **kwargs):
    delta_n = kwargs['delta_n'] if 'delta_n' in kwargs else -2
    if power:
        cv = power/delta_n
    elif r:
        cv = 1.0/r
    else:
        cv = c

    if ec:
        k = ec - 1.0
    else:
        k = cc

    if profile is Spherical:
        prf = Spherical(c=cv)
    elif profile is Conic:
        prf = Conic(c=cv, cc=k)
    else:
        if k == 0.0:
            prf = Spherical(c=cv)
        else:
            prf = Conic(c=cv, cc=k)

    m = Surface(profile=prf, interact_mode='reflect', max_ap=sd,
                delta_n=delta_n, **kwargs)
    me = Mirror(m, sd=sd)
    return [[m, None, None, 1, -1]], [me]


def create_lens(power=0., bending=0., th=None, sd=1., med=None):
    if med is None:
        med = Glass()
    rndx = med.rindex('d')
    cv1 = power/(2*(rndx - 1))
    cv2 = -power/(2*(rndx - 1))
    s1 = Surface(profile=Spherical(c=cv1), max_ap=sd, delta_n=(rndx - 1))
    s2 = Surface(profile=Spherical(c=cv2), max_ap=sd, delta_n=(1 - rndx))
    if th is None:
        th = sd/5
    g = Gap(t=th, med=med)
    le = Element(s1, s2, g, sd=sd)
    return [[s1, g, None, rndx, 1], [s2, None, None, 1, 1]], [le]


def create_dummy_plane(sd=1., **kwargs):
    s = Surface(**kwargs)
    se = DummyInterface(s, sd=sd)
    return [[s, None, None, 1, +1]], [se]


def create_air_gap(t=0., ref_ifc=None):
    g = Gap(t=t)
    ag = AirGap(g, ref_ifc)
    return g, ag


def create_from_file(filename, **kwargs):
    opm = cmds.open_model(filename)
    sm = opm.seq_model
    osp = opm.optical_spec
    fod = osp.parax_data.fod
    if 'power' in kwargs:
        desired_power = kwargs['power']
        cur_power = fod.power
        scale_factor = desired_power/cur_power
        sm.apply_scale_factor(scale_factor)
    seq = [list(node) for node in sm.path(start=1, stop=-1)]
    ele = opm.ele_model.elements[:-2]
    return seq, ele


class Element():
    clut = rgbt.RGBTable(filename='red_blue64.csv',
                         data_range=[10.0, 100.])

    label_format = 'E{}'

    def __init__(self, s1, s2, g, tfrm=None, idx=0, idx2=1, sd=1.,
                 label='Lens'):
        self.label = label
        if tfrm is not None:
            self.tfrm = tfrm
        else:
            self.tfrm = (np.identity(3), np.array([0., 0., 0.]))
        self.s1 = s1
        self.s1_indx = idx
        self.s2 = s2
        self.s2_indx = idx2
        self.gap = g
        self.medium_name = self.gap.medium.name()
        self._sd = sd
        self.flat1 = None
        self.flat2 = None
        self.render_color = self.calc_render_color()
        self.handles = {}
        self.actions = {}

    @property
    def sd(self):
        return self._sd

    @sd.setter
    def sd(self, semidiam):
        self._sd = semidiam
        self.edge_extent = (-semidiam, semidiam)

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['parent']
        del attrs['tfrm']
        del attrs['s1']
        del attrs['s2']
        del attrs['gap']
        del attrs['handles']
        del attrs['actions']
        return attrs

    def __str__(self):
        fmt = 'Element: {!r}, {!r}, t={:.4f}, sd={:.4f}, glass: {}'
        return fmt.format(self.s1.profile, self.s2.profile, self.gap.thi,
                          self.sd, self.gap.medium.name())

    def sync_to_restore(self, ele_model, surfs, gaps, tfrms):
        # when restoring, we want to use the stored indices to look up the
        # new object instances
        self.parent = ele_model
        self.tfrm = tfrms[self.s1_indx]
        self.s1 = surfs[self.s1_indx]
        self.gap = gaps[self.s1_indx]
        self.s2 = surfs[self.s2_indx]
        if not hasattr(self, 'medium_name'):
            self.medium_name = self.gap.medium.name()

    def sync_to_update(self, seq_model):
        # when updating, we want to use the stored object instances to get the
        # current indices into the interface list (e.g. to handle insertion and
        # deletion of interfaces)
        self.s1_indx = seq_model.ifcs.index(self.s1)
        self.s2_indx = seq_model.ifcs.index(self.s2)
        self.render_color = self.calc_render_color()

    def reference_interface(self):
        return self.s1

    def reference_idx(self):
        return self.s1_indx

    def interface_list(self):
        return [self.s1, self.s2]

    def gap_list(self):
        return [self.gap]

    def get_bending(self):
        cv1 = self.s1.profile_cv
        cv2 = self.s2.profile_cv
        delta_cv = cv1 - cv2
        bending = 0.
        if delta_cv != 0.0:
            bending = (cv1 + cv2)/delta_cv
        return bending

    def set_bending(self, bending):
        cv1 = self.s1.profile_cv
        cv2 = self.s2.profile_cv
        delta_cv = cv1 - cv2
        cv2_new = 0.5*(bending - 1.)*delta_cv
        cv1_new = bending*delta_cv - cv2_new
        self.s1.profile_cv = cv1_new
        self.s2.profile_cv = cv2_new

    def update_size(self):
        extents = np.union1d(self.s1.get_y_aperture_extent(),
                             self.s2.get_y_aperture_extent())
        self.edge_extent = (extents[0], extents[-1])
        self.sd = max(self.s1.surface_od(), self.s2.surface_od())
        return self.sd

    def calc_render_color(self):
        try:
            gc = float(self.gap.medium.glass_code())
        except AttributeError:
            return (255, 255, 255, 64)  # white
        else:
            # set element color based on V-number
            indx, vnbr = glass_decode(gc)
            dsg, rgb = gp.find_glass_designation(indx, vnbr)
#            rgb = Element.clut.get_color(vnbr)
            return rgb

    def compute_flat(self, s):
        ca = s.surface_od()
        if (1.0 - ca/self.sd) >= 0.05:
            flat = ca
        else:
            flat = None
        return flat

    def extent(self):
        if hasattr(self, 'edge_extent'):
            return self.edge_extent
        else:
            return (-self.sd, self.sd)

    def render_shape(self):
        if self.s1.profile_cv < 0.0:
            self.flat1 = self.compute_flat(self.s1)
        poly = self.s1.full_profile(self.extent(), self.flat1)
        if self.s2.profile_cv > 0.0:
            self.flat2 = self.compute_flat(self.s2)
        poly2 = self.s2.full_profile(self.extent(), self.flat2, -1)
        for p in poly2:
            p[0] += self.gap.thi
        poly += poly2
        poly.append(poly[0])
        return poly

    def render_handles(self, opt_model):
        self.handles = {}
        ifcs_gbl_tfrms = opt_model.seq_model.gbl_tfrms

        shape = self.render_shape()
        self.handles['shape'] = GraphicsHandle(shape, self.tfrm, 'polygon')

        extent = self.extent()
        if self.flat1 is not None:
            extent_s1 = self.flat1,
        else:
            extent_s1 = extent
        poly_s1 = self.s1.full_profile(extent_s1, None)
        gh1 = GraphicsHandle(poly_s1, ifcs_gbl_tfrms[self.s1_indx], 'polyline')
        self.handles['s1_profile'] = gh1

        if self.flat2 is not None:
            extent_s2 = self.flat2,
        else:
            extent_s2 = extent
        poly_s2 = self.s2.full_profile(extent_s2, None, -1)
        gh2 = GraphicsHandle(poly_s2, ifcs_gbl_tfrms[self.s2_indx], 'polyline')
        self.handles['s2_profile'] = gh2

        poly_sd_upr = []
        poly_sd_upr.append([poly_s1[-1][0], extent[1]])
        poly_sd_upr.append([poly_s2[0][0]+self.gap.thi, extent[1]])
        self.handles['sd_upr'] = GraphicsHandle(poly_sd_upr, self.tfrm,
                                                'polyline')

        poly_sd_lwr = []
        poly_sd_lwr.append([poly_s2[-1][0]+self.gap.thi, extent[0]])
        poly_sd_lwr.append([poly_s1[0][0], extent[0]])
        self.handles['sd_lwr'] = GraphicsHandle(poly_sd_lwr, self.tfrm,
                                                'polyline')

        poly_ct = []
        poly_ct.append([0., 0.])
        poly_ct.append([self.gap.thi, 0.])
        self.handles['ct'] = GraphicsHandle(poly_ct, self.tfrm, 'polyline')

        return self.handles

    def handle_actions(self):
        self.actions = {}

        shape_actions = {}
        shape_actions['pt'] = BendAction(self)
        shape_actions['y'] = AttrAction(self, 'sd')
        self.actions['shape'] = shape_actions

        s1_prof_actions = {}
        s1_prof_actions['pt'] = SagAction(self.s1)
        self.actions['s1_profile'] = s1_prof_actions

        s2_prof_actions = {}
        s2_prof_actions['pt'] = SagAction(self.s2)
        self.actions['s2_profile'] = s2_prof_actions

        sd_upr_action = {}
        sd_upr_action['y'] = AttrAction(self, 'sd')
        self.actions['sd_upr'] = sd_upr_action

        sd_lwr_action = {}
        sd_lwr_action['y'] = AttrAction(self, 'sd')
        self.actions['sd_lwr'] = sd_lwr_action

        ct_action = {}
        ct_action['x'] = AttrAction(self.gap, 'thi')
        self.actions['ct'] = ct_action

        return self.actions


class Mirror():

    label_format = 'M{}'

    def __init__(self, ifc, tfrm=None, idx=0, sd=1., thi=None, z_dir=1.0,
                 label='Mirror'):
        self.label = label
#        self.render_color = (192, 192, 192, 112)
        self.render_color = (158, 158, 158, 64)
#        self.render_color = (64, 64, 64, 64)
        if tfrm is not None:
            self.tfrm = tfrm
        else:
            self.tfrm = (np.identity(3), np.array([0., 0., 0.]))
        self.s = ifc
        self.s_indx = idx
        self.z_dir = z_dir
        self.sd = sd
        self.flat = None
        self.thi = thi
        self.medium_name = 'Mirror'
        self.handles = {}
        self.actions = {}

    def get_thi(self):
        thi = self.thi
        if self.thi is None:
            thi = 0.05*self.sd
        return thi

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['parent']
        del attrs['tfrm']
        del attrs['s']
        del attrs['handles']
        del attrs['actions']
        return attrs

    def __str__(self):
        thi = self.get_thi()
        fmt = 'Mirror: {!r}, t={:.4f}, sd={:.4f}'
        return fmt.format(self.s.profile, thi, self.sd)

    def sync_to_restore(self, ele_model, surfs, gaps, tfrms):
        self.parent = ele_model
        self.tfrm = tfrms[self.s_indx]
        self.s = surfs[self.s_indx]
        if not hasattr(self, 'medium_name'):
            self.medium_name = 'Mirror'

    def reference_interface(self):
        return self.s

    def reference_idx(self):
        return self.s_indx

    def interface_list(self):
        return [self.s]

    def gap_list(self):
        return []

    def sync_to_update(self, seq_model):
        self.s_indx = seq_model.ifcs.index(self.s)

    def update_size(self):
        self.edge_extent = self.s.get_y_aperture_extent()
        self.sd = self.s.surface_od()
        return self.sd

    def extent(self):
        if hasattr(self, 'edge_extent'):
            return self.edge_extent
        else:
            self.edge_extent = self.s.get_y_aperture_extent()
            return self.edge_extent

    def render_shape(self):
        poly = self.s.full_profile(self.extent(), self.flat)
        poly2 = self.s.full_profile(self.extent(), self.flat, -1)

        thi = self.get_thi()
        offset = thi*self.z_dir

        for p in poly2:
            p[0] += offset
        poly += poly2
        poly.append(poly[0])
        return poly

    def render_handles(self, opt_model):
        self.handles = {}
        ifcs_gbl_tfrms = opt_model.seq_model.gbl_tfrms

        self.handles['shape'] = GraphicsHandle(self.render_shape(), self.tfrm,
                                               'polygon')

        poly = self.s.full_profile(self.extent(), None)
        self.handles['s_profile'] = GraphicsHandle(poly,
                                                   ifcs_gbl_tfrms[self.s_indx],
                                                   'polyline')

        thi = self.get_thi()
        offset = thi*self.z_dir

        poly_sd_upr = []
        poly_sd_upr.append(poly[-1])
        poly_sd_upr.append([poly[-1][0]+offset, poly[-1][1]])
        self.handles['sd_upr'] = GraphicsHandle(poly_sd_upr, self.tfrm,
                                                'polyline')

        poly_sd_lwr = []
        poly_sd_lwr.append(poly[0])
        poly_sd_lwr.append([poly[0][0]+offset, poly[0][1]])
        self.handles['sd_lwr'] = GraphicsHandle(poly_sd_lwr, self.tfrm,
                                                'polyline')

        return self.handles

    def handle_actions(self):
        self.actions = {}

        shape_actions = {}
        shape_actions['pt'] = SagAction(self.s)
        self.actions['shape'] = shape_actions

        s_prof_actions = {}
        s_prof_actions['pt'] = SagAction(self.s)
        self.actions['s_profile'] = s_prof_actions

        sd_upr_action = {}
        sd_upr_action['y'] = AttrAction(self, 'edge_extent[1]')
        self.actions['sd_upr'] = sd_upr_action

        sd_lwr_action = {}
        sd_lwr_action['y'] = AttrAction(self, 'edge_extent[0]')
        self.actions['sd_lwr'] = sd_lwr_action

        return self.actions


class ThinElement():

    label_format = 'TL{}'

    def __init__(self, ifc, tfrm=None, idx=0, sd=None, label='ThinLens'):
        self.label = label
        self.render_color = (192, 192, 192)
        if tfrm is not None:
            self.tfrm = tfrm
        else:
            self.tfrm = (np.identity(3), np.array([0., 0., 0.]))
        self.intrfc = ifc
        self.intrfc_indx = idx
        self.medium_name = 'Thin Element'
        if sd is not None:
            self.sd = sd
        else:
            self.sd = ifc.max_aperture
        self.handles = {}
        self.actions = {}

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['parent']
        del attrs['tfrm']
        del attrs['intrfc']
        del attrs['handles']
        del attrs['actions']
        return attrs

    def __str__(self):
        return str(self.intrfc)

    def sync_to_restore(self, ele_model, surfs, gaps, tfrms):
        self.parent = ele_model
        self.tfrm = tfrms[self.intrfc_indx]
        self.intrfc = surfs[self.intrfc_indx]
        if not hasattr(self, 'medium_name'):
            self.medium_name = 'Thin Element'

    def reference_interface(self):
        return self.intrfc

    def reference_idx(self):
        return self.intrfc_indx

    def interface_list(self):
        return [self.intrfc]

    def gap_list(self):
        return []

    def sync_to_update(self, seq_model):
        self.intrfc_indx = seq_model.ifcs.index(self.intrfc)

    def update_size(self):
        self.sd = self.intrfc.surface_od()
        return self.sd

    def render_shape(self):
        poly = self.intrfc.full_profile((-self.sd, self.sd))
        return poly

    def render_handles(self, opt_model):
        self.handles = {}
        shape = self.render_shape()
        self.handles['shape'] = GraphicsHandle(shape, self.tfrm, 'polygon')
        return self.handles

    def handle_actions(self):
        self.actions = {}
        return self.actions


class DummyInterface():

    label_format = 'D{}'

    def __init__(self, ifc, idx=0, sd=None, tfrm=None, label='DummyInterface'):
        self.label = label
        self.render_color = (192, 192, 192)
        if tfrm is not None:
            self.tfrm = tfrm
        else:
            self.tfrm = (np.identity(3), np.array([0., 0., 0.]))
        self.ref_ifc = ifc
        self.idx = idx
        self.medium_name = 'Interface'
        if sd is not None:
            self.sd = sd
        else:
            self.sd = ifc.max_aperture
        self.handles = {}
        self.actions = {}

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['parent']
        del attrs['tfrm']
        del attrs['ref_ifc']
        del attrs['handles']
        del attrs['actions']
        return attrs

    def __str__(self):
        return str(self.ref_ifc)

    def sync_to_restore(self, ele_model, surfs, gaps, tfrms):
        self.parent = ele_model
        self.tfrm = tfrms[self.idx]
        self.ref_ifc = surfs[self.idx]
        if not hasattr(self, 'medium_name'):
            self.medium_name = 'Interface'

    def reference_interface(self):
        return self.ref_ifc

    def reference_idx(self):
        return self.idx

    def interface_list(self):
        return [self.ref_ifc]

    def gap_list(self):
        return []

    def sync_to_update(self, seq_model):
        self.idx = seq_model.ifcs.index(self.ref_ifc)

    def update_size(self):
        self.sd = self.ref_ifc.surface_od()
        return self.sd

    def render_shape(self):
        poly = self.ref_ifc.full_profile((-self.sd, self.sd))
        return poly

    def render_handles(self, opt_model):
        self.handles = {}

        self.handles['shape'] = GraphicsHandle(self.render_shape(), self.tfrm,
                                               'polyline')

        return self.handles

    def handle_actions(self):
        self.actions = {}

        def get_adj_spaces():
            seq_model = self.parent.opt_model.seq_model
            if self.idx > 0:
                before = seq_model.gaps[self.idx-1].thi
            else:
                before = None
            if self.idx < seq_model.get_num_surfaces() - 1:
                after = seq_model.gaps[self.idx].thi
            else:
                after = None
            return (before, after)

        def set_adj_spaces(cur_value, change):
            seq_model = self.parent.opt_model.seq_model
            if cur_value[0] is not None:
                seq_model.gaps[self.idx-1].thi = cur_value[0] + change
            if cur_value[1] is not None:
                seq_model.gaps[self.idx].thi = cur_value[1] - change

        slide_action = {}
        slide_action['x'] = Action(get_adj_spaces, set_adj_spaces)
        self.actions['shape'] = slide_action

        return self.actions


class AirGap():

    label_format = 'AirGap {}'

    def __init__(self, g, ref_ifc, idx=0, tfrm=None, label='AirGap'):
        if tfrm is not None:
            self.tfrm = tfrm
        else:
            self.tfrm = (np.identity(3), np.array([0., 0., 0.]))
        self.label = label
        self.render_color = (237, 243, 254, 64)  # light blue
        self.gap = g
        self.medium_name = self.gap.medium.name()
        self.ref_ifc = ref_ifc
        self.idx = idx
        self.handles = {}
        self.actions = {}

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['parent']
        del attrs['tfrm']
        del attrs['gap']
        del attrs['ref_ifc']
        del attrs['handles']
        del attrs['actions']
        return attrs

    def __str__(self):
        return str(self.gap)

    def sync_to_restore(self, ele_model, surfs, gaps, tfrms):
        self.parent = ele_model
        self.gap = gaps[self.idx]
        self.ref_ifc = surfs[self.idx]
        self.tfrm = tfrms[self.idx]
        if not hasattr(self, 'render_color'):
            self.render_color = (237, 243, 254, 64)  # light blue
        if not hasattr(self, 'medium_name'):
            self.medium_name = self.gap.medium.name()

    def reference_interface(self):
        return self.ref_ifc

    def reference_idx(self):
        return self.idx

    def interface_list(self):
        return []

    def gap_list(self):
        return [self.gap]

    def sync_to_update(self, seq_model):
        self.idx = seq_model.gaps.index(self.gap)

    def update_size(self):
        pass

    def render_handles(self, opt_model):
        self.handles = {}

        poly_ct = []
        poly_ct.append([0., 0.])
        poly_ct.append([self.gap.thi, 0.])
        self.handles['ct'] = GraphicsHandle(poly_ct, self.tfrm, 'polyline')

        return self.handles

    def handle_actions(self):
        self.actions = {}

        ct_action = {}
        ct_action['x'] = AttrAction(self.gap, 'thi')
        self.actions['ct'] = ct_action

        return self.actions


class ElementModel:

    def __init__(self, opt_model):
        self.opt_model = opt_model
        self.elements = []
        self.ifcs_dict = {}
        self.gap_dict = {}

    def reset(self):
        self.__init__()

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['ifcs_dict']
        del attrs['gap_dict']
        return attrs

    def sync_to_restore(self, opt_model):
        self.opt_model = opt_model
        seq_model = opt_model.seq_model
        surfs = seq_model.ifcs
        gaps = seq_model.gaps
        tfrms = seq_model.compute_global_coords(1)

        self.ifcs_dict = {}
        self.gap_dict = {}

        # special processing for older models
        self.airgaps_from_sequence(seq_model, tfrms)
        self.add_dummy_interface_at_image(seq_model, tfrms)

        for i, e in enumerate(self.elements):
            e.sync_to_restore(self, surfs, gaps, tfrms)
            for ifc in e.interface_list():
                self.ifcs_dict[ifc] = e
            for g in e.gap_list():
                self.gap_dict[g] = e
            if not hasattr(e, 'label'):
                e.label = e.label_format.format(i+1)
        self.sequence_elements()
        self.relabel_airgaps()
#        self.list_elements()

    def elements_from_sequence(self, seq_model):
        """ generate an element list from a sequential model """

        # if there are elements in the list already, just return
        if len(self.elements) > 0:
            return

        num_elements = 0
        tfrms = seq_model.compute_global_coords(1)
        for i, g in enumerate(seq_model.gaps):
            s1 = seq_model.ifcs[i]
            tfrm = tfrms[i]
            if g.medium.name().lower() == 'air':
                if i > 0:
                    self.process_airgap(seq_model, i, g, s1, tfrm,
                                        num_elements, add_ele=True)
            else:  # a non-air medium
                # handle buried mirror, e.g. prism or Mangin mirror
                if s1.interact_mode == 'reflect':
                    gp = seq_model.gaps[i-1]
                    if gp.medium.name().lower() == g.medium.name().lower():
                        continue

                s2 = seq_model.ifcs[i+1]
                sd = max(s1.surface_od(), s2.surface_od())
                e = Element(s1, s2, g, sd=sd, tfrm=tfrm, idx=i, idx2=i+1)
                num_elements += 1
                e.label = Element.label_format.format(num_elements)
                self.add_element(e)

        self.add_dummy_interface_at_image(seq_model, tfrms)

        self.relabel_airgaps()

    def process_airgap(self, seq_model, i, g, s, tfrm, num_ele, add_ele=True):
        if s.interact_mode == 'reflect' and add_ele:
            sd = s.surface_od()
            z_dir = seq_model.z_dir[i]
            m = Mirror(s, sd=sd, tfrm=tfrm, idx=i, z_dir=z_dir)
            num_ele += 1
            m.label = Mirror.label_format.format(num_ele)
            self.add_element(m)
        elif s.interact_mode == 'transmit':
            add_dummy = False
            if i == 0:
                add_dummy = True  # add dummy for the object
                dummy_label = 'Object'
            else:  # i > 0
                gp = seq_model.gaps[i-1]
                if gp.medium.name().lower() == 'air':
                    add_dummy = True
                    if seq_model.stop_surface == i:
                        dummy_label = 'Aperture Stop'
                    else:
                        dummy_label = DummyInterface.label_format.format(i)
            if add_dummy:
                tfrm = tfrm
                sd = s.surface_od()
                di = DummyInterface(s, sd=sd, tfrm=tfrm, idx=i)
                di.label = dummy_label
                self.add_element(di)
        elif isinstance(s, thinlens.ThinLens) and add_ele:
            te = ThinElement(s, tfrm=tfrm, idx=i)
            num_ele += 1
            te.label = ThinElement.label_format.format(num_ele)
            self.add_element(te)

        # add an AirGap
        ag = AirGap(g, s, idx=i, tfrm=tfrm)
        self.add_element(ag)

    def airgaps_from_sequence(self, seq_model, tfrms):
        """ add airgaps and dummy interfaces to an older version model """
        for e in self.elements:
            if isinstance(e, AirGap):
                return  # found an AirGap, model probably OK

        num_elements = 0
        seq_model = self.opt_model.seq_model
        for i, g in enumerate(seq_model.gaps):
            if g.medium.name().lower() == 'air':
                if i > 0:
                    s = seq_model.ifcs[i]
                    tfrm = tfrms[i]
                    self.process_airgap(seq_model, i, g, s, tfrm,
                                        num_elements, add_ele=False)

    def add_dummy_interface_at_image(self, seq_model, tfrms):
        if len(self.elements) and self.elements[-1].label == 'Image':
            return

        s = seq_model.ifcs[-1]
        idx = seq_model.get_num_surfaces() - 1
        di = DummyInterface(s, sd=s.surface_od(), tfrm=tfrms[-1], idx=idx)
        di.label = 'Image'
        self.add_element(di)

    def update_model(self):
        seq_model = self.opt_model.seq_model
        tfrms = seq_model.compute_global_coords(1)
        for e in self.elements:
            e.update_size()
            e.sync_to_update(seq_model)
            intrfc = e.reference_interface()
            try:
                i = seq_model.ifcs.index(intrfc)
            except ValueError:
                print("Interface {} not found".format(intrfc.lbl))
            else:
                e.tfrm = tfrms[i]
        self.sequence_elements()

    def sequence_elements(self):
        """ Sort elements in order of reference interfaces in seq_model """
        seq_model = self.opt_model.seq_model
        self.elements.sort(key=lambda e:
                           seq_model.ifcs.index(e.reference_interface()))

    def relabel_airgaps(self):
        for i, e in enumerate(self.elements):
            if isinstance(e, AirGap):
                eb = self.elements[i-1].label
                ea = self.elements[i+1].label
                e.label = AirGap.label_format.format(eb + '-' + ea)

    def add_element(self, e):
        e.parent = self
        self.elements.append(e)
        for ifc in e.interface_list():
            self.ifcs_dict[ifc] = e
        for g in e.gap_list():
            self.gap_dict[g] = e

    def remove_element(self, e):
        for ifc in e.interface_list():
            self.ifcs_dict.pop(ifc)
        for g in e.gap_list():
            self.gap_dict.pop(g)
        e.parent = None
        self.elements.remove(e)

    def get_num_elements(self):
        return len(self.elements)

    def list_elements(self):
        for i, ele in enumerate(self.elements):
            print("%d: %s (%s): %s" %
                  (i, ele.label, type(ele).__name__, ele))

    def element_type(self, i):
        return type(self.elements[i]).__name__



""" Package encompassing optical modeling and calculations

    The ``rayoptics.optical`` subpackage provides core classes and functions
    for optical modeling and basic analyses. These include:

        - Sequential optical model, :mod:`~.sequential`
        - Element based model, :mod:`~.elements`
        - Optical specifications, :mod:`~.opticalspec`
        - Paraxial and third order calculations, :mod:`~.firstorder`,
          :mod:`~.thirdorder`
        - Primitive and higher level ray tracing, :mod:`~.raytrace`,
          :mod:`~.trace`

    The overall optical model is managed by the :class:`~.OpticalModel` class
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Static optical system layout

.. Created on Sat Jun 30 10:15:37 2018

.. codeauthor: Michael J. Hayford
"""

from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Polygon

import numpy as np

from rayoptics.gui import layout
from rayoptics.gui import util
from rayoptics.util.rgb2mpl import rgb2mpl, backgrnd_color


class LensLayoutFigure(Figure):
    """  Static optical system layout

    Attributes:
        opt_model: parent optical model
        do_draw_frame: if True, draw frame around system layout
        oversize_factor: what fraction to oversize the system bounding box
        offset_factor: how much to draw rays before first surface
        do_draw_rays: if True, draw edge rays
    """
    def __init__(self, opt_model,
                 do_draw_frame=False,
                 oversize_factor=0.05,
                 offset_factor=0.05,
                 do_draw_rays=True,
                 **kwargs):
        self.opt_model = opt_model
        self.linewidth = 0.5
        self.do_draw_frame = do_draw_frame
        self.do_draw_rays = do_draw_rays
        self.oversize_factor = oversize_factor
        self.offset_factor = offset_factor

        Figure.__init__(self, **kwargs)

        self.set_facecolor(backgrnd_color)

        ele_model = self.opt_model.ele_model
        if len(ele_model.elements) == 0:
            ele_model.elements_from_sequence(self.opt_model.seq_model)

        self.update_data()

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        self.artists = []

        self.ele_shapes = self.create_element_model(self.opt_model.ele_model)
        self.ele_bbox = self.update_patches(self.ele_shapes)

        if self.do_draw_rays:
            offset = self.offset_factor
            self.ray_shapes = self.create_ray_model(offset_factor=offset)
            self.ray_bbox = self.update_patches(self.ray_shapes)

            concat_bbox = np.concatenate((self.ele_bbox, self.ray_bbox))
            self.sys_bbox = util.bbox_from_poly(concat_bbox)
        else:
            self.sys_bbox = self.ele_bbox

        return self

    def system_length(self):
        osp = self.opt_model.optical_spec
        img_dist = abs(osp.parax_data[2].img_dist)
        ele_length = self.ele_bbox[1][0] - self.ele_bbox[0][0]
        return ele_length+img_dist

    def update_patches(self, shapes):
        bbox_list = []
        for shape in shapes:
            poly, bbox = shape[0](shape[1])
            self.artists.append(poly)
            if len(bbox_list) == 0:
                bbox_list = bbox
            else:
                bbox_list = np.vstack((bbox_list, bbox))
        bbox = util.bbox_from_poly(bbox_list)
        return bbox

    def create_element_model(self, ele_model):
        elements = []
        for e in ele_model.elements:
            oe = layout.create_optical_element(self.opt_model, e)
            handles = oe.update_shape(self)
            if 'shape' in handles:
                elements.append((self.update_element_shape, oe))
        return elements

    def update_element_shape(self, oe):
        handles = oe.update_shape(self)
        return handles['shape']

    def create_ray_model(self, start_surf=1, offset_factor=0.05):
        start_offset = offset_factor*self.system_length()

        ray_bundles = []
        fov = self.opt_model.optical_spec.field_of_view
        wvl = self.opt_model.seq_model.central_wavelength()
        for i, fld in enumerate(fov.fields):
            fld_label = fov.index_labels[i]
            rb = layout.RayBundle(self.opt_model, fld, fld_label,
                                  wvl, start_offset)
            ray_bundles.append((self.update_ray_fan_shape, rb))
        return ray_bundles

    def create_polygon(self, poly, rgb_color, **kwargs):
        p = Polygon(poly, closed=True, fc=rgb2mpl(rgb_color),
                    ec='black', **kwargs)
        p.set_linewidth(self.linewidth)
        return p

    def create_polyline(self, poly, **kwargs):
        x = poly.T[0]
        y = poly.T[1]
        p = Line2D(x, y, linewidth=2)
        return p

    def update_ray_fan_shape(self, rb):
        handles = rb.update_shape(self)
        return handles['shape']

    def scale_bounds(self, oversize_factor):
        bbox = util.scale_bounds(self.sys_bbox, oversize_factor)
        self.ax.set_xlim(bbox[0][0], bbox[1][0])
        self.ax.set_ylim(bbox[0][1], bbox[1][1])

    def draw_frame(self, do_draw_frame):
        if do_draw_frame:
            self.ax.set_axis_on()
            self.tight_layout(pad=0.)
        else:
            self.ax.set_axis_off()
            self.tight_layout(pad=0.)

    def plot(self):
        try:
            self.ax.cla()
        except AttributeError:
            self.ax = self.add_subplot(1, 1, 1, aspect=1.0)

        for p in self.artists:
            if isinstance(p, Line2D):
                self.ax.add_line(p)
            elif isinstance(p, Patch):
                self.ax.add_patch(p)
            else:
                self.ax.add_artist(p)

        self.scale_bounds(self.oversize_factor)
        self.draw_frame(self.do_draw_frame)
        self.ax.set_facecolor(backgrnd_color)

        self.canvas.draw()

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

.. Created on Tue Nov 27 21:26:08 2018

.. codeauthor: Michael J. Hayford
"""

import logging
from collections import namedtuple

from PyQt5 import QtCore
from PyQt5.QtCore import QDate, QSize, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QCheckBox, QComboBox, QDateTimeEdit, QAction,
                             QDialog, QGroupBox, QHBoxLayout, QLabel,
                             QLineEdit, QListView, QListWidget, QDockWidget,
                             QListWidgetItem, QPushButton, QSpinBox,
                             QStackedWidget, QVBoxLayout, QFormLayout, QWidget)

from rayoptics.optical.model_enums import PupilType
from rayoptics.optical.model_enums import FieldType
from rayoptics.optical.model_enums import DimensionType


PanelInfo = namedtuple('PanelInfo', ['dock', 'panel_widget', 'menu_action'])

panels = {}


def create_dock_windows(gui_app):
    panels['wavelengths'] = create_dock_widget(gui_app,
                                               'wavelengths', 'Wavelengths',
                                               SpectrumWavelengthsPanel, False)
    panels['aperture'] = create_dock_widget(gui_app,
                                            'aperture', 'Aperture',
                                            AperturePanel, False)
    panels['field'] = create_dock_widget(gui_app,
                                         'field', 'Field of View',
                                         FieldOfViewPanel, False)
    panels['system'] = create_dock_widget(gui_app,
                                          'system', 'System Info',
                                          SystemSpecPanel, False)


def create_dock_widget(gui_app, item_key, label, panel, state):
    dock = QDockWidget(item_key, gui_app)
    panel_widget = panel(gui_app, dock)
    dock.setWidget(panel_widget)
    menu_action = create_menu_action(gui_app, item_key, label, state)
    gui_app.addDockWidget(Qt.RightDockWidgetArea, dock)
    dock.setVisible(state)
    return PanelInfo(dock, panel_widget, menu_action)


def update_dock_windows(gui_app):
    for pi in panels.values():
        if pi.menu_action.isChecked():
            pi.panel_widget.update(gui_app.app_manager.model)


def create_menu_action(gui_app, item_key, label, state=False):
    menu_action = QAction(label, gui_app, checkable=True)
    menu_action.setChecked(state)
    menu_action.triggered.connect(lambda state:
                                  togglePanel(gui_app, state, item_key))
    return menu_action


def togglePanel(gui_app, state, item_key):
    panel_info = panels[item_key]
    panel_info.dock.setVisible(state)
    if state:
        panel_info.panel_widget.update(gui_app.app_manager.model)


class UIWidget():
    def __init__(self, gui_app, widgetLabel, widgetClass,
                 rootEvalStr, get_eval_str, set_eval_str=None):
        self.gui_app = gui_app
        self.rootEvalStr = rootEvalStr
        self.get_eval_str = get_eval_str
        if set_eval_str is not None:
            self.set_eval_str = set_eval_str
        else:
            self.set_eval_str = get_eval_str + '={}'
#            self.set_eval_str = get_eval_str + '="{:s}"'
        self.widgetLabel = widgetLabel
        self.widget = widgetClass()

    def get_root_object(self):
        root = self.gui_app.app_manager.model
        if len(self.rootEvalStr) == 0:
            return root
        else:
            root_eval_str = ('root' + self.rootEvalStr)
            try:
                root = eval(root_eval_str)
            except IndexError:
                return root
            else:
                return root

    def data(self):
        root = self.get_root_object()
        eval_str = 'root' + self.get_eval_str
        try:
            val = eval(eval_str)
            return val
        except IndexError:
            return ''
        except TypeError:
            print('Data type error: ', eval_str, val)
            return ''

    def setData(self, value):
        root = self.get_root_object()
        exec_str = ('root' + self.set_eval_str).format(value)
        try:
            exec(exec_str)
            return True
        except IndexError:
            return False
        except SyntaxError:
            logging.info('Syntax error: "%s"', value)
            return False

    def updateData(self):
        """ push widget data to backend """
        pass

    def refresh(self):
        """ push backend data to widget """
        pass


class TextFieldWidget(UIWidget):
    def __init__(self, gui_app, widgetLabel, widgetClass,
                 rootEvalStr, get_eval_str, valueFormat, set_eval_str=None):
        super().__init__(gui_app, widgetLabel, widgetClass, rootEvalStr,
                         get_eval_str, set_eval_str)
        self.valueFormat = valueFormat
        self.widget.editingFinished.connect(self.updateData)
        self.widget.setAlignment(QtCore.Qt.AlignLeft)

    def updateData(self):
        """ push widget data to backend """
        self.setData(self.widget.text())
        self.gui_app.refresh_gui()

    def refresh(self):
        """ push backend data to widget """
        valStr = self.valueFormat.format(self.data())
        self.widget.setText(valStr)


class ChoiceWidget(UIWidget):
    def __init__(self, gui_app, widgetLabel, widgetClass,
                 rootEvalStr, get_eval_str, combo_items, set_eval_str=None):
        super().__init__(gui_app, widgetLabel, widgetClass, rootEvalStr,
                         get_eval_str, set_eval_str)
        for item in combo_items:
            self.widget.addItem(item)
        self.widget.currentIndexChanged.connect(self.updateData)

    def updateData(self):
        """ push widget data to backend """
        self.setData(self.widget.currentIndex())
        self.gui_app.refresh_gui()

    def refresh(self):
        """ push backend data to widget """
        self.widget.setCurrentIndex(self.data())


class SpectrumWavelengthsPanel(QWidget):
    rootEvalStr = '.optical_spec.spectral_region'
    evalStr = '.central_wvl', '.wavelengths[0]', '.wavelengths[-1]'

    def __init__(self, gui_app, parent=None):
        super().__init__(parent)

        self.ctrl_wvl_edit = TextFieldWidget(
            gui_app, 'central', QLineEdit,
            SpectrumWavelengthsPanel.rootEvalStr,
            SpectrumWavelengthsPanel.evalStr[0],
            '{:7.1f}')

        self.red_wvl_edit = TextFieldWidget(
            gui_app, 'red', QLineEdit,
            SpectrumWavelengthsPanel.rootEvalStr,
            SpectrumWavelengthsPanel.evalStr[1],
            '{:7.1f}')

        self.blue_wvl_edit = TextFieldWidget(
            gui_app, 'blue', QLineEdit,
            SpectrumWavelengthsPanel.rootEvalStr,
            SpectrumWavelengthsPanel.evalStr[2],
            '{:7.1f}')

        wavlnsLayout = QFormLayout()
        wavlnsLayout.addRow('central', self.ctrl_wvl_edit.widget)
        wavlnsLayout.addRow('red', self.red_wvl_edit.widget)
        wavlnsLayout.addRow('blue', self.blue_wvl_edit.widget)

        self.achroCheckBox = QCheckBox("Achromatic")
        wavlnsLayout.addRow('', self.achroCheckBox)

        self.setLayout(wavlnsLayout)

    def update(self, opt_model):
        """ push backend data to widgets """
        num_wvls = len(opt_model.optical_spec.spectral_region.wavelengths)
        if num_wvls == 1:
            self.ctrl_wvl_edit.refresh()
            self.red_wvl_edit.widget.setText('')
            self.blue_wvl_edit.widget.setText('')
        elif num_wvls == 2:
            self.ctrl_wvl_edit.widget.setText('')
            self.red_wvl_edit.refresh()
            self.blue_wvl_edit.refresh()
        elif num_wvls > 2:
            self.ctrl_wvl_edit.refresh()
            self.red_wvl_edit.refresh()
            self.blue_wvl_edit.refresh()


class AperturePanel(QWidget):
    rootEvalStr = '.optical_spec.pupil'
    evalStr = '.value', '.get_pupil_type()'
    comboItems = ["Ent Pupil Diam", "Object NA", "F/#", "NA"]
    set_combo_str = '.mutate_pupil_type(PupilType({}))'

    def __init__(self, gui_app, parent=None):
        super().__init__(parent)

        apertureLayout = QFormLayout()

        self.aperture_combo = ChoiceWidget(gui_app, 'pupil_type', QComboBox,
                                           AperturePanel.rootEvalStr,
                                           AperturePanel.evalStr[1],
                                           AperturePanel.comboItems,
                                           set_eval_str=
                                           AperturePanel.set_combo_str)
        apertureLayout.addRow('Type', self.aperture_combo.widget)

        self.aperture_edit = TextFieldWidget(gui_app, 'value', QLineEdit,
                                             AperturePanel.rootEvalStr,
                                             AperturePanel.evalStr[0],
                                             '{:12.5f}')
        apertureLayout.addRow('value', self.aperture_edit.widget)

        self.setLayout(apertureLayout)

    def update(self, opt_model):
        """ push backend data to widgets """
        self.aperture_combo.refresh()
        self.aperture_edit.refresh()


class FieldOfViewPanel(QWidget):
    rootEvalStr = '.optical_spec.field_of_view'
    evalStr = '.value', '.get_field_type()'
    comboItems = ["Object Angle", "Object Height", "Image Height"]
    set_combo_str = '.mutate_field_type(FieldType({}))'

    def __init__(self, gui_app, parent=None):
        super().__init__(parent)

        fieldLayout = QFormLayout()

        self.field_combo = ChoiceWidget(gui_app, 'field_type', QComboBox,
                                        FieldOfViewPanel.rootEvalStr,
                                        FieldOfViewPanel.evalStr[1],
                                        FieldOfViewPanel.comboItems,
                                        set_eval_str=
                                        FieldOfViewPanel.set_combo_str)
        fieldLayout.addRow('Type', self.field_combo.widget)

        self.field_edit = TextFieldWidget(gui_app, 'value', QLineEdit,
                                          FieldOfViewPanel.rootEvalStr,
                                          FieldOfViewPanel.evalStr[0],
                                          '{:12.5f}')
        fieldLayout.addRow('value', self.field_edit.widget)

        self.setLayout(fieldLayout)

    def update(self, opt_model):
        """ push backend data to widgets """
        self.field_combo.refresh()
        self.field_edit.refresh()


class SystemSpecPanel(QWidget):
    rootEvalStr = '.system_spec'
    evalStr = ('.dimensions.value', '.title', '.initials', '.temperature',
               '.pressure')
    comboItems = ["mm", "cm", "m", "inches"]
    set_combo_str = '.dimensions=DimensionType({})'
    set_text_str = ".title='{:s}'", ".initials='{:s}'"

    def __init__(self, gui_app, parent=None):
        super().__init__(parent)

        systemLayout = QFormLayout()

        self.dimension_combo = ChoiceWidget(gui_app, 'dimensions', QComboBox,
                                            SystemSpecPanel.rootEvalStr,
                                            SystemSpecPanel.evalStr[0],
                                            SystemSpecPanel.comboItems,
                                            set_eval_str=
                                            SystemSpecPanel.set_combo_str)
        systemLayout.addRow('system units', self.dimension_combo.widget)

        self.title_edit = TextFieldWidget(gui_app, 'title', QLineEdit,
                                          SystemSpecPanel.rootEvalStr,
                                          SystemSpecPanel.evalStr[1],
                                          '{:s}',
                                          set_eval_str=
                                          SystemSpecPanel.set_text_str[0])
        systemLayout.addRow('title', self.title_edit.widget)

        self.initials_edit = TextFieldWidget(gui_app, 'initials', QLineEdit,
                                             SystemSpecPanel.rootEvalStr,
                                             SystemSpecPanel.evalStr[2],
                                             '{:s}',
                                             set_eval_str=
                                             SystemSpecPanel.set_text_str[1])
        systemLayout.addRow('initials', self.initials_edit.widget)

        self.temp_edit = TextFieldWidget(gui_app, 'temperature', QLineEdit,
                                         SystemSpecPanel.rootEvalStr,
                                         SystemSpecPanel.evalStr[3],
                                         '{:7.1f}')
        systemLayout.addRow('temperature', self.temp_edit.widget)

        self.pressure_edit = TextFieldWidget(gui_app, 'pressure', QLineEdit,
                                             SystemSpecPanel.rootEvalStr,
                                             SystemSpecPanel.evalStr[4],
                                             '{:9.3f}')
        systemLayout.addRow('pressure', self.pressure_edit.widget)

        self.setLayout(systemLayout)

    def update(self, opt_model):
        """ push backend data to widgets """
        self.dimension_combo.refresh()
        self.title_edit.refresh()
        self.initials_edit.refresh()
        self.temp_edit.refresh()
        self.pressure_edit.refresh()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" utility functions for gui functions

.. Created on Wed Oct 16 15:27:40 2019

.. codeauthor: Michael J. Hayford
"""


from collections import namedtuple
import numpy as np

GUIHandle = namedtuple('GUIHandle', ['poly', 'bbox'])
""" tuple grouping together graphics entity and bounding box

    Attributes:
        poly: poly entity for underlying graphics system (e.g. mpl)
        bbox: bounding box for poly
"""


def transform_ray_seg(poly, r, tfrm):
    rot, trns = tfrm
    p = rot.dot(r.p) + trns
    poly.append([p[2], p[1]])


def bbox_from_poly(poly):
    minx, miny = np.min(poly, axis=0)
    maxx, maxy = np.max(poly, axis=0)
    return np.array([[minx, miny], [maxx, maxy]])


def scale_bounds(bbox, oversize_factor):
    inc_x = oversize_factor*(bbox[1][0] - bbox[0][0])
    inc_y = oversize_factor*(bbox[1][1] - bbox[0][1])
    incr = max(inc_x, inc_y)
    return np.array([[bbox[0][0]-incr, bbox[0][1]-incr],
                     [bbox[1][0]+incr, bbox[1][1]+incr]])


def transform_poly(tfrm, poly):
    coord_flip = np.array([[0., 1.], [1., 0.]])

    poly = np.matmul(coord_flip, poly.T)
    poly = np.matmul(tfrm[0][1:, 1:], poly).T

    t = np.array([tfrm[1][1], tfrm[1][2]])
    poly += t

    # flip coordinates back to 2D plot coordinates, +y points up
    poly = np.matmul(poly, coord_flip)
    bbox = bbox_from_poly(poly)
    return poly, bbox


def inv_transform_poly(tfrm, poly):
    coord_flip = np.array([[0., 1.], [1., 0.]])
    try:
        poly = np.matmul(coord_flip, poly.T)
    except TypeError:
        print(poly)

    t = np.array([tfrm[1][1], tfrm[1][2]])
    poly -= t

    poly = np.matmul(tfrm[0][1:, 1:], poly).T

    # flip coordinates back to 2D plot coordinates, +y points up
    poly = np.matmul(poly, coord_flip)
    return poly


def fit_data_range(x_data, margin=0.05, range_trunc=0.25, **kwargs):
    x_min = min(0., min(x_data))
    x_max = max(0., max(x_data))
    x_range = x_max - x_min
    if x_range != 0.0 and len(x_data) > 2:
        x1_min = min(0., min(x_data[1:]))
        x1_max = max(0., max(x_data[1:]))
        x1_range = x1_max - x1_min
        if abs(x1_range/x_range) < range_trunc:
            x_min = x1_min
            x_max = x1_max
            x_range = x1_range

    if x_range > 0.:
        x_margin = margin*x_range
    else:
        x_margin = 0.01
    return x_min-x_margin, x_max+x_margin



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Useful transforms for processing sequential models

.. Created on Fri Feb  9 10:09:58 2018

.. codeauthor: Michael J. Hayford
"""

import numpy as np


def forward_transform(s1, zdist, s2):
    """ generate transform rotation and translation from
        s1 coords to s2 coords """

    # calculate origin of s2 wrt to s1
    t_orig = np.array([0., 0., zdist])
    r_after_s1 = r_before_s2 = None
    if s1.decenter:
        # get transformation info after s1
        r_after_s1, t_after_s1 = s1.decenter.tform_after_surf()
        t_orig += t_after_s1

    if s2.decenter:
        # get transformation info before s2
        r_before_s2, t_before_s2 = s2.decenter.tform_before_surf()
        t_orig += t_before_s2

    r_cascade = np.identity(3)
    if r_after_s1 is not None:
        # rotate the origin of s2 around s1 "after" transformation
        t_orig = r_after_s1.dot(t_orig)
        r_cascade = r_after_s1
        if r_before_s2 is not None:
            r_cascade = r_after_s1.dot(r_before_s2)
    elif r_before_s2 is not None:
        r_cascade = r_before_s2

    return r_cascade, t_orig


def reverse_transform(s1, zdist, s2):
    """ generate transform rotation and translation from
        s2 coords to s1 coords """

    # calculate origin of s2 wrt to s1
    t_orig = np.array([0., 0., zdist])
    r_after_s1 = r_before_s2 = None
    if s1.decenter:
        # get transformation info after s1
        r_after_s1, t_after_s1 = s1.decenter.tform_after_surf()
        t_orig += t_after_s1

    if s2.decenter:
        # get transformation info before s2
        r_before_s2, t_before_s2 = s2.decenter.tform_before_surf()
        t_orig += t_before_s2

    # going in reverse direction so negate translation
    t_orig = -t_orig

    r_cascade = np.identity(3)
    if r_before_s2 is not None:
        # rotate the origin of s1 around s2 "before" transformation
        r_cascade = r_before_s2.transpose()
        t_orig = r_cascade.dot(t_orig)
        if r_after_s1 is not None:
            r_cascade = r_cascade.dot(r_after_s1.transpose())
    elif r_after_s1 is not None:
        r_cascade = r_after_s1.transpose()

    return r_cascade, t_orig


def cascade_transform(r_prev, t_prev, r_seg, t_seg):
    """ take the seg transform and cascade it with the prev transform """
    return r_prev.dot(r_seg), r_prev.dot(t_seg) + t_prev


def transfer_coords(r_seg, t_seg, pt_s1, dir_s1):
    """ take p and d in s1 coords of seg and transfer them to s2 coords """
    rt = r_seg.transpose()
    return rt.dot(pt_s1 - t_seg), rt.dot(dir_s1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
"""

.. Created on Mon Apr  2 19:20:27 2018

.. codeauthor: Michael J. Hayford
"""

from collections import namedtuple

import numpy as np
from matplotlib.figure import Figure

from rayoptics.optical.model_constants import ht, slp, aoi
from rayoptics.optical.elements import (create_thinlens, create_mirror,
                                        create_lens)

from rayoptics.util.misc_math import distance_sqr_2d, perpendicular_distance_2d


SelectInfo = namedtuple('SelectInfo', ['artist', 'info'])
""" tuple grouping together an artist and info returned from contains(event)

    Attributes:
        artist: the artist
        info: a dictionary of artist specific details of selection
"""


def create_parax_design_commands(fig):
    cmds = []
    # Select an existing point
    cmds.append(('Select', (fig.select_point, (), {})))
    # Add thin lens
    cmds.append(('Add Thin Lens',
                 (fig.add_point, (), {'factory': create_thinlens})))
    # Add lens
    cmds.append(('Add Lens', (fig.add_point, (), {'factory': create_lens})))
    # Add mirror
    cmds.append(('Add Mirror',
                 (fig.add_point, (), {'factory': create_mirror})))

    return cmds


class EditableLine:
    def __init__(self, line):
        self.line = line
        self.press = None
        self.is_hilited = False
        self.markers, = self.line.axes.plot([], [], 'rs')
        self.pick_radius_sqr = self.line.get_pickradius()**2

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.line.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.line.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.line.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def process_hit_location(self, event, hit_list, prt=False):
        xdata, ydata = self.line.get_data()
        # create a list of vertex coordinates in the model coordinate system
        line_hits = [[xdata[i], ydata[i]] for i in hit_list]
        # transform the list into display coordinates
        dsp_hits = self.line.axes.transData.transform(line_hits)
        hit_pt = [event.x, event.y]

        hit_vertex = None
        hit_edge = None
        min_hit_dist = 1e10
        min_perp_dist = 1e10
        for i, pt in enumerate(dsp_hits):
            # consider hit point as a vertex
            hit_dist = distance_sqr_2d(pt, hit_pt)
            if hit_dist < min_hit_dist:
                min_hit_dist = hit_dist
                if hit_dist < self.pick_radius_sqr:
                    hit_vertex = hit_list[i]

            # consider hit point on an edge
            nxt = hit_list[i]+1
            try:
                next_pt = np.array([xdata[nxt], ydata[nxt]])
            except IndexError:
                continue
            else:
                next_dsp_pt = self.line.axes.transData.transform(next_pt)
                perp_dist = perpendicular_distance_2d(pt, hit_pt, next_dsp_pt)
                if perp_dist < min_perp_dist:
                    min_perp_dist = perp_dist
#                    if perp_dist < self.pick_radius_sqr:
#                    print("perp_dist", hit_list[i], perp_dist)
                    hit_edge = hit_list[i]
        v = None
        if hit_vertex is not None:
            v = (hit_vertex, event.xdata, event.ydata,
                 xdata[hit_vertex], ydata[hit_vertex], 's')
#            print("vertex selected %d: event x=%f y=%f data x=%f y=%f" %
#                  (hit_vertex,
#                   event.xdata, event.ydata,
#                   xdata[hit_vertex], ydata[hit_vertex]))
        e = None
        if hit_edge is not None:
            h = hit_edge
#            print("edge selected", hit_list, min_hit_dist,
#                  xdata[h:h+2], ydata[h:h+2])
            e = (hit_edge, event.xdata, event.ydata,
                 xdata[h:h+2], ydata[h:h+2], '')

        return (v, e)

    def on_press(self, event):
        # on button press we will see if the mouse is over us and store
        #  some data
#        print("eline.on_press")
        if event.inaxes != self.line.axes:
            return

        contains, attrd = self.line.contains(event)
        if not contains:
            return

        hit_list = attrd['ind']
        v, e = self.process_hit_location(event, hit_list)

        self.press = v, e

    def on_motion(self, event):
        'on motion we will highlight a vertex of edge if the mouse is over it '
        if event.inaxes != self.line.axes:
            return

        contains, props = self.line.contains(event)
        if not contains:
            self.markers.set_data([], [])
            self.line.figure.canvas.draw()
            return

#        print("eline.on_motion")
        hit_list = props['ind']
        v, e = self.process_hit_location(event, hit_list)

        if v:
            hit_vertex, x_hit, y_hit, x_data, y_data, mkr = v
        elif e:
            hit_vertex, x_hit, y_hit, x_data, y_data, mkr = e

        if v or e:
            self.markers.set_xdata(x_data)
            self.markers.set_ydata(y_data)
            self.markers.set_marker(mkr)
            self.markers.set_linestyle('solid')
            self.line.figure.canvas.draw()

    def on_release(self, event):
        'on release we reset the press data'
        self.press = None
        self.line.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.line.figure.canvas.mpl_disconnect(self.cidpress)
        self.line.figure.canvas.mpl_disconnect(self.cidrelease)
        self.line.figure.canvas.mpl_disconnect(self.cidmotion)


class ParaxialDesignFigure(Figure):

    def __init__(self, opt_model, refresh_gui, dgm_type, **kwargs):
        self.actions = {}
        self.opt_model = opt_model
        self.parax_model = opt_model.parax_model
        self.refresh_gui = refresh_gui
        self.setup_dgm_type(dgm_type)
        self.skip_build = False

        Figure.__init__(self, **kwargs)

        self.vertex = None
        self.update_data()
        self.select_point()

    def setup_dgm_type(self, dgm_type):
        if dgm_type == 'ht':
            self.type_sel = ht
            self.data_slice = slice(0, None)
#            self.data_slice = slice(1, None)
            self.x_label = r'$\overline{y}$'
            self.y_label = 'y'
            self.apply_data = self.parax_model.apply_ht_dgm_data
            self.header = r'$y-\overline{y}$ Diagram'
        elif dgm_type == 'slp':
            self.type_sel = slp
            self.data_slice = slice(0, -1)
            self.x_label = r'$\overline{\omega}$'
            self.y_label = r'$\omega$'
            self.apply_data = self.parax_model.apply_slope_dgm_data
            self.header = r'$\omega-\overline{\omega}$ Diagram'

    def refresh(self):
        self.update_data()
        self.plot()

    def update_data(self):
        if not self.skip_build:
            self.parax_model.build_lens()
        self.skip_build = False
        return self

    def fit_data_range(self, x_data, margin=0.05, range_trunc=0.25):
        x_min = min(0., min(x_data))
        x_max = max(0., max(x_data))
        x_range = x_max - x_min
        if x_range != 0.0 and len(x_data) > 2:
            x1_min = min(0., min(x_data[1:]))
            x1_max = max(0., max(x_data[1:]))
            x1_range = x1_max - x1_min
            if abs(x1_range/x_range) < range_trunc:
                x_min = x1_min
                x_max = x1_max
                x_range = x1_range

        if x_range > 0.:
            x_margin = margin*x_range
        else:
            x_margin = 0.01
        return x_min-x_margin, x_max+x_margin

    def plot(self):
        self.clf()
        self.ax = self.add_subplot(1, 1, 1)
        self.ax.axvline(0, c='black', lw=1)
        self.ax.axhline(0, c='black', lw=1)
        self.ax.set_title(self.header, pad=10.0, fontsize=18)

        x_data = []
        y_data = []
        for x, y in zip(self.parax_model.pr,
                        self.parax_model.ax):
            x_data.append(x[self.type_sel])
            y_data.append(y[self.type_sel])

        x_min, x_max = self.fit_data_range(x_data)
        y_min, y_max = self.fit_data_range(y_data)

        self.line, = self.ax.plot(x_data,
                                  y_data,
                                  marker='s', picker=6)
        self.ax.set_xlabel(self.x_label)
        self.ax.set_ylabel(self.y_label)
        self.ax.set_xlim(x_min, x_max)
        self.ax.set_ylim(y_min, y_max)

        xl, xr = self.ax.get_xlim()
        yl, yu = self.ax.get_ylim()
        xlb, xrb = self.ax.get_xbound()
        ylb, yub = self.ax.get_ybound()

        self.eline = EditableLine(self.line)
        self.eline.connect()
        self.canvas.mpl_connect('pick_event', self.on_pick)
        self.canvas.mpl_connect('button_press_event', self.on_press)
        self.canvas.mpl_connect('button_release_event', self.on_release)
        self.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.canvas.draw()

        return self

    def select_point(self, **kwargs):
        def on_press_select_point(pdfig, press):
            hit_vertex, x_hit, y_hit, x_data, y_data, mkr = press
            pdfig.vertex = hit_vertex + pdfig.data_slice.start
        self.actions['press'] = on_press_select_point

        def on_drag_select_point(pdfig, event):
            pdfig.apply_data(pdfig.vertex, (event.xdata, event.ydata))
            pdfig.parax_model.paraxial_lens_to_seq_model()
            pdfig.skip_build = True
            pdfig.refresh_gui()
        self.actions['drag'] = on_drag_select_point

        def on_release_select_point(pdfig, event):
            pdfig.apply_data(pdfig.vertex, (event.xdata, event.ydata))
            pdfig.parax_model.paraxial_lens_to_seq_model()
            pdfig.skip_build = True
            pdfig.refresh_gui()
            pdfig.vertex = None
        self.actions['release'] = on_release_select_point

        self.move_action = None

    def add_point(self, factory=None, **kwargs):
        def on_press_add_point(pdfig, press):
            hit_vertex, x_hit, y_hit, x_data, y_data, mkr = press
            new_vertex = hit_vertex + pdfig.data_slice.start
#            print("add_point:", new_vertex)
            pdfig.parax_model.add_object(new_vertex, (x_hit, y_hit),
                                         pdfig.type_sel, factory, 'transmit')
            new_vertex += 1
            pdfig.parax_model.paraxial_lens_to_seq_model()
            pdfig.skip_build = True
            pdfig.refresh_gui()
        self.actions['press'] = on_press_add_point

    def on_press(self, event):
#        print("on_press")
        hit, props = self.line.contains(event)
        if hit:
            if self.eline.press is None:
                self.eline.on_press(event)
            if self.eline.press:
                v, e = self.eline.press
                if v:
                    self.actions['press'](self, v)
                    hit_vertex, x_hit, y_hit, x_data, y_data, mkr = v
                    self.vertex = hit_vertex + self.data_slice.start
#                    print("vertex selected", hit_vertex)
                elif e:
                    hit_edge, x_hit, y_hit, x_data, y_data, mkr = e
                    self.actions['press'](self, e)
#                    print("edge selected", e[0])

#            print('on_press', event.button, event.x, event.y,
#                  event.xdata, event.ydata, event.key,
#                  len(hit_list), hit_list)

    def on_motion(self, event):
        if self.vertex:
            self.actions['drag'](self, event)

    def on_release(self, event):
        'on release we reset the press data'
        if self.vertex:
            self.actions['release'](self, event)

    def on_pick(self, event):
        line = event.artist
        me = event.mouseevent
        xdata, ydata = line.get_data()
        ind = event.ind
        id = line.get_gid()
#        print("on_pick", id, ind, xdata[ind], ydata[ind], me.name, me.x, me.y,
#              me.button, me.key, me.xdata, me.ydata, me.dblclick)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Support for reading CODE V .seq files

.. codeauthor: Michael J. Hayford
"""

import re


def tokenize_command(cmd):
    tkns = re.findall(r"[^'\"\s]\S*|\".+?\"|'.+?'", cmd)
    tokens = []
    for t in tkns:
        if t[:1] == '"':
            tokens.append(t.strip('"'))
        elif t[:1] == "'":
            tokens.append(t.strip("'"))
        else:
            tokens.append(t)
    return tokens


def next_line(it):
    ln = next(it)
    contIndex = ln.rfind('&')
    if contIndex >= 0:
        return ln[:contIndex] + next_line(it)
    else:
        return ln


def strip_comments(textLine):
    commentIndex = textLine.find('!')
    if commentIndex >= 0:
        return textLine[:commentIndex]
    else:
        return textLine


def read_seq_buffer(inputLines, tokenize=True):
    inputs = []
    it = iter(inputLines)
    while True:
        try:
            ln = next_line(it)
            ln = strip_comments(ln)
            lnList = ln.split(';')
            for line in lnList:
                line = line.strip()
                if len(line) > 0:
                    if tokenize:
                        cmd = tokenize_command(line)
                        inputs.append(cmd)
                    else:
                        inputs.append(line)

        except StopIteration:
            break

    return inputs


def read_seq_file(filename, tokenize=True):
    file = open(filename, 'r')
    inpt = file.read()
    file.close()
    inputLines = inpt.splitlines()
    return read_seq_buffer(inputLines, tokenize)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" First order paraxial design space

.. Created on Sat Mar 31 21:14:42 2018

.. codeauthor: Michael J. Hayford
"""

from collections import namedtuple
import numpy as np

from rayoptics.optical.model_constants import ht, slp, aoi
from rayoptics.optical.model_constants import pwr, tau, indx, rmd
import rayoptics.optical.model_constants as mc
import rayoptics.optical.firstorder as fo
from rayoptics.optical.gap import Gap
from rayoptics.optical.surface import Surface


def bbox_from_poly(poly):
    minx, miny = np.min(poly, axis=0)
    maxx, maxy = np.max(poly, axis=0)
    return np.array([[minx, miny], [maxx, maxy]])


ParaxData = namedtuple('ParaxData', ['ht', 'slp', 'aoi'])
""" paraxial ray data at an interface

    Attributes:
        ht: height at interface
        slp: n*slope
        aoi: n*angle of incidence
"""

ParaxSys = namedtuple('ParaxSys', ['pwr', 'tau', 'indx', 'rmd'])
""" paraxial lens data at an interface

    Attributes:
        pwr: power
        tau: reduced distance, t/n
        indx: refractive index (n)
        rmd: refract mode
"""


class ParaxialModel():
    def __init__(self, opt_model):
        self.opt_model = opt_model
        self.seq_model = opt_model.seq_model
        self.sys = None
        self.ax = []
        self.pr = []
        self.opt_inv = 1.0

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['seq_model']
        del attrs['parax_data']
        return attrs

    def sync_to_restore(self, opt_model):
        self.opt_model = opt_model
        self.seq_model = opt_model.seq_model

    def update_model(self):
        self.parax_data = self.opt_model.optical_spec.parax_data
        self.build_lens()

    def build_lens(self):
        # rebuild the `sys` description from the seq_model path
        self.sys = sys = self.seq_path_to_paraxial_lens(self.seq_model.path())

        # precalculate the reduced forms of the paraxial axial and chief rays
        ax_ray, pr_ray, fod = self.parax_data
        self.opt_inv = fod.opt_inv

        self.ax = []
        self.pr = []
        for i in range(0, len(sys)):
            n = sys[i][indx]
            self.ax.append([ax_ray[i][ht], n*ax_ray[i][slp], n*ax_ray[i][aoi]])
            self.pr.append([pr_ray[i][ht], n*pr_ray[i][slp], n*pr_ray[i][aoi]])

    def add_node(self, node, new_vertex, type_sel, interact_mode):
        """ Add a node in the paraxial data structures """
        ns = self.seq_model.get_num_surfaces()
        if node >= ns - 1:
            node = ns - 2
        n = self.sys[node][indx]
        new_node = node + 1
        self.sys.insert(new_node, [0.0, 0.0, n, interact_mode])

        ax_node = [0.0, 0.0, 0.0]
        ax_node[type_sel] = new_vertex[1]
        self.ax.insert(new_node, ax_node)

        pr_node = [0.0, 0.0, 0.0]
        pr_node[type_sel] = new_vertex[0]
        self.pr.insert(new_node, pr_node)

        if type_sel == ht:
            self.apply_ht_dgm_data(new_node, new_vertex=new_vertex)
        elif type_sel == slp:
            self.apply_slope_dgm_data(new_node, new_vertex=new_vertex)

        if interact_mode == 'reflect':
            for i in range(new_node, len(self.sys)):
                self.sys[i][indx] = -self.sys[i][indx]

        return new_node

    def assign_object_to_node(self, node, factory, **inputs):
        """ create a new element from `factory` and replace `node` with it """

        # extract optical properties of node
        n = self.sys[node][indx]
        power = self.sys[node][pwr]
        thi = n*self.sys[node][tau]
        sd = abs(self.ax[node][ht]) + abs(self.pr[node][ht])

        # create an element with the node's properties
        seq, ele = factory(power=power, sd=sd)

        n_before = self.sys[node-1][indx]
        thi_before = n_before*self.sys[node-1][tau]
        self.seq_model.gaps[node-1].thi = thi_before

        # insert the path sequence and elements into the
        #  sequential and element models
        args = seq, ele
        kwargs = {'idx': node-1, 't': thi, **inputs}
        self.opt_model.insert_ifc_gp_ele(*args, **kwargs)

        path_stop = node + len(seq)
        inserted_seq = list(self.seq_model.path(start=node-1, stop=path_stop))
        sys_seq = self.seq_path_to_paraxial_lens(inserted_seq[1:])
        pp_info = self.compute_principle_points(node, inserted_seq)
        self.replace_node_with_seq(node, sys_seq, pp_info)

        return args, kwargs

    def compute_principle_points(self, node, seq):
        n_0 = seq[0][mc.Indx]
        z_dir_before = seq[0][mc.Zdir]
        n_k = seq[-1][mc.Indx]
        z_dir_k = seq[-1][mc.Zdir]
        path = [[Surface(), Gap(), None, n_0, z_dir_before]]
        path.extend(seq[1:])
        pp_info = fo.compute_principle_points(iter(path),
                                              n_0=z_dir_before*n_0,
                                              n_k=z_dir_k*n_k)
        return pp_info

    def replace_node_with_seq(self, node, sys_seq, pp_info):
        """ replaces the data at node with sys_seq """
        efl, pp1, ppk, ffl, bfl = pp_info[2]
        self.sys[node-1][tau] -= pp1/self.sys[node-1][indx]
        sys_seq[-1][tau] = self.sys[node][tau] - ppk/sys_seq[-1][indx]

        self.delete_node(node)
        for ss in sys_seq:
            self.sys.insert(node, ss)
            self.ax.insert(node, [0.0, 0.0, 0.0])
            self.pr.insert(node, [0.0, 0.0, 0.0])
            node += 1
        self.paraxial_trace()

    def get_object_for_node(self, node):
        ''' basic 1:1 relationship between seq and parax model sequences '''
        ifc = self.seq_model.ifcs[node]
        ele = self.opt_model.ele_model.ifcs_dict[ifc]
        args = [[ifc, None, None, 1, 1]], [ele]
        kwargs = {'idx': node}
        return args, kwargs

    def add_object(self, surf, new_vertex, type_sel, factory, interact_mode):
        new_surf = self.add_node(surf, new_vertex, type_sel, interact_mode)
        self.assign_object_to_node(new_surf, factory)

    def delete_node(self, surf):
        """ delete the node at position surf """
        del self.sys[surf]
        del self.ax[surf]
        del self.pr[surf]

    def apply_ht_dgm_data(self, surf, new_vertex=None):
        """ This routine calculates all data dependent on the input
            height coordinates (y,ybar) at surface surf.
        """
        sys = self.sys
        ax_ray = self.ax
        pr_ray = self.pr
        opt_inv = self.opt_inv

        if new_vertex is not None:
            pr_ray[surf][ht] = new_vertex[0]
            ax_ray[surf][ht] = new_vertex[1]

        nsm1 = len(sys) - 1
        if surf == 0:
            surf += 1

        p = surf - 1
        c = surf

        sys[p][tau] = ((ax_ray[p][ht]*pr_ray[c][ht] -
                        ax_ray[c][ht]*pr_ray[p][ht]) / opt_inv)
        ax_ray[p][slp] = (ax_ray[c][ht] - ax_ray[p][ht])/sys[p][tau]
        pr_ray[p][slp] = (pr_ray[c][ht] - pr_ray[p][ht])/sys[p][tau]

        if (surf > 1):
            p2 = surf - 2
            sys[p][pwr] = ((ax_ray[p2][slp]*pr_ray[p][slp] -
                            ax_ray[p][slp]*pr_ray[p2][slp])
                           / opt_inv)

        if (surf < nsm1):
            s = surf + 1

            sys[c][tau] = (ax_ray[c][ht]*pr_ray[s][ht] -
                           ax_ray[s][ht]*pr_ray[c][ht])/opt_inv
            ax_ray[c][slp] = (ax_ray[s][ht] - ax_ray[c][ht])/sys[c][tau]
            pr_ray[c][slp] = (pr_ray[s][ht] - pr_ray[c][ht])/sys[c][tau]
            sys[c][pwr] = (ax_ray[p][slp]*pr_ray[c][slp] -
                           ax_ray[c][slp]*pr_ray[p][slp])/opt_inv

            if s < nsm1:
                sys[s][pwr] = (ax_ray[c][slp]*pr_ray[s][slp] -
                               ax_ray[s][slp]*pr_ray[c][slp])/opt_inv

        else:
            ax_ray[c][slp] = ax_ray[p][slp]
            pr_ray[c][slp] = pr_ray[p][slp]
            sys[c][pwr] = 0
            sys[c][tau] = 0

    def apply_slope_dgm_data(self, surf, new_vertex=None):
        """ This routine calculates all data dependent on the input
            slope coordinates (nu,nubar) at surface surf.
        """
        sys = self.sys
        ax_ray = self.ax
        pr_ray = self.pr
        opt_inv = self.opt_inv

        if new_vertex is not None:
            pr_ray[surf][slp] = new_vertex[0]
            ax_ray[surf][slp] = new_vertex[1]

        nsm1 = len(sys) - 1
        if nsm1 == 0:
            p = 0
            c = 1
            ax_ray[c][ht] = ax_ray[p][slp]*sys[p][tau] + ax_ray[p][ht]
            pr_ray[c][ht] = pr_ray[p][slp]*sys[p][tau] + pr_ray[p][ht]

        else:
            if (surf == 0):
                surf += 1

            p = surf - 1
            c = surf

            sys[c][pwr] = (ax_ray[p][slp]*pr_ray[c][slp] -
                           ax_ray[c][slp]*pr_ray[p][slp])/opt_inv
            ax_ray[c][ht] = (ax_ray[p][slp] - ax_ray[c][slp])/sys[c][pwr]
            pr_ray[c][ht] = (pr_ray[p][slp] - pr_ray[c][slp])/sys[c][pwr]
            sys[p][tau] = (ax_ray[p][ht]*pr_ray[c][ht] -
                           ax_ray[c][ht]*pr_ray[p][ht])/opt_inv

            if (surf < nsm1):
                s = surf + 1
                ax_ray[s][ht] = ax_ray[c][slp]*sys[c][tau] + ax_ray[c][ht]
                pr_ray[s][ht] = pr_ray[c][slp]*sys[c][tau] + pr_ray[c][ht]

    # ParaxTrace() - This routine performs a paraxial raytrace from object
    #                (surface 0) to image.  The last operation is a
    #                transfer to the image surface.
    def paraxial_trace(self):
        """ regenerate paraxial axial and chief rays from power and reduced
            distance
        """
        sys = self.sys
        ax_ray = self.ax
        pr_ray = self.pr

        nsm1 = len(sys) - 1

        # Transfer from object
        c = 0
        s = 1
        ax_ray[s][ht] = ax_ray[c][ht] + sys[c][tau]*ax_ray[c][slp]
        pr_ray[s][ht] = pr_ray[c][ht] + sys[c][tau]*pr_ray[c][slp]

        for i in range(1, len(sys)):

            p = c
            c = s

            # Refraction
            ax_ray[c][slp] = ax_ray[p][slp] - ax_ray[c][ht]*sys[c][pwr]
            pr_ray[c][slp] = pr_ray[p][slp] - pr_ray[c][ht]*sys[c][pwr]

            # Transfer
            if (i < nsm1):
                s += 1
                ax_ray[s][ht] = ax_ray[c][ht] + sys[c][tau]*ax_ray[c][slp]
                pr_ray[s][ht] = pr_ray[c][ht] + sys[c][tau]*pr_ray[c][slp]

    def list_lens(self):
        """ list the paraxial axial and chief rays, and power, reduced distance
        """
        sys = self.sys
        ax_ray = self.ax
        pr_ray = self.pr

        print("       ax_ray_ht    ax_ray_slp")
        for i in range(0, len(ax_ray)):
            print("{:2}: {:12.5g}  {:12.6g}".format(i, ax_ray[i][ht],
                                                    ax_ray[i][slp]))

        print("\n       pr_ray_ht    pr_ray_slp")
        for i in range(0, len(pr_ray)):
            print("{:2}: {:12.5g}  {:12.6g}".format(i, pr_ray[i][ht],
                                                    pr_ray[i][slp]))

        print("\n            power           tau        index    type")
        for i in range(0, len(sys)):
            print("{:2}: {:13.7g}  {:12.5g} {:12.5f}    {}".format(
                i, sys[i][pwr], sys[i][tau], sys[i][indx], sys[i][rmd]))

    def seq_path_to_paraxial_lens(self, path):
        """ returns lists of power, reduced thickness, signed index and refract
            mode
        """
        sys = []
        for i, sg in enumerate(path):
            n_after = sg[mc.Indx] if sg[mc.Zdir] > 0 else -sg[mc.Indx]
            imode = sg[mc.Intfc].interact_mode
            power = sg[mc.Intfc].optical_power
            if sg[mc.Gap]:
                tau = sg[mc.Gap].thi/n_after
                sys.append([power, tau, n_after, imode])
            else:
                sys.append([power, 0.0, n_after, imode])
        return sys

    def paraxial_lens_to_seq_model(self):
        """ Applies a paraxial lens spec (power, reduced distance) to the model
        """
        sys = self.sys
        ax_ray = self.ax
        n_before = sys[0][indx]
        slp_before = self.ax[0][slp]
        for i, sg in enumerate(self.seq_model.path()):
            if sg[mc.Gap]:
                n_after = sys[i][indx]
                slp_after = ax_ray[i][slp]
                sg[mc.Gap].thi = n_after*sys[i][tau]

                sg[mc.Intfc].set_optical_power(sys[i][pwr], n_before, n_after)
                sg[mc.Intfc].from_first_order(slp_before, slp_after,
                                              ax_ray[i][ht])

                n_before = n_after
                slp_before = slp_after

    def pwr_slope_solve(self, ray, surf, slp_new):
        p = ray[surf-1]
        c = ray[surf]
        pwr = (p[slp] - slp_new)/c[ht]
        return pwr

    def pwr_ht_solve(self, ray, surf, ht_new):
        sys = self.sys
        p = ray[surf-1]
        c = ray[surf]
        slp_new = (ht_new - c[ht])/sys[surf][tau]
        pwr = (p[slp] - slp_new)/ht_new
        return pwr

    def thi_ht_solve(self, ray, surf, ht_new):
        c = ray[surf]
        thi = (ht_new - c[ht])/c[slp]
        return thi

    def apply_data(self, vertex, lcl_pt):
        ray = self.ray
        p = ray[vertex-1]
        c = ray[vertex]
        c_slp_new = (lcl_pt[1] - c[ht])/lcl_pt[0]
        pwr = (p[slp] - c_slp_new)/c[ht]
        self.opt_model.seq_model.ifcs[vertex].optical_power = pwr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" functions to calculate etendue and imager parameters

.. Created on Tue Aug  6 18:08:16 2019

.. codeauthor: Michael J. Hayford
"""
import math

from rayoptics.util import dict2d


obj_img_set = ['object', 'image']
fld_ape_set = ['field', 'aperture']
fld_labels = ['height', 'angle']
ap_labels = ['pupil', 'NA', 'f/#']


def na2slp(na, n=1.0):
    """ convert numerical aperture to slope """
    return n*math.tan(math.asin(na/n))


def slp2na(slp, n=1.0):
    """ convert a ray slope to numerical aperture """
    return n*math.sin(math.atan(slp/n))


def ang2slp(ang):
    """ convert an angle in degrees to a slope """
    return math.tan(math.radians(ang))


def slp2ang(slp):
    """ convert a slope to an angle in degrees """
    return math.degrees(math.atan(slp))


def get_aperture_from_slope(imager, slope, n=1):
    if slope == 0:
        return 0, 0, 0
    fno = -1/(2*slope)
    na = slp2na(slope, n=n)
    pupil = imager.f/fno
    return na, fno, pupil


def get_slope_from_aperture(imager, input_cell, n=1):
    if 'NA' in input_cell:
        na = input_cell['NA']
        slope = na2slp(na, n=n)
    elif 'f/#' in input_cell:
        fno = input_cell['f/#']
        slope = -1/(2*fno)
    elif 'pupil' in input_cell:
        pupil = input_cell['pupil']
        if imager.f is None or imager.f == 0:
            slope = 0
        else:
            slope = -pupil/(2*imager.f)
    else:
        slope = None
    return slope


def calc_aperture_from_input(imager, input_cell, n=1):
    slope = get_slope_from_aperture(imager, input_cell, n=n)
    na, fno, pupil = get_aperture_from_slope(imager, slope, n=n)
    return na, fno, pupil


def do_etendue_via_imager(conj_type, imager, etendue_inputs, etendue_grid,
                          n_0=1, n_k=1):
    imager_inputs = {}
    li = dict2d.num_items_by_type(etendue_inputs, fld_ape_set, obj_img_set)
    if li['field'] == 1:
        row = dict2d.row(etendue_inputs, 'field')
        obj_img_key = 'object' if len(row['object']) else 'image'
        do_field_via_imager(conj_type, imager, etendue_inputs, obj_img_key,
                            etendue_grid, n_0=n_0, n_k=n_k)

    if li['aperture'] == 1:
        row = dict2d.row(etendue_inputs, 'aperture')
        obj_img_key = 'object' if len(row['object']) else 'image'
        do_aperture_via_imager(conj_type, imager, etendue_inputs, obj_img_key,
                               etendue_grid)

    if li['field'] == 2 or li['aperture'] == 2:
        fld_ape_key = 'field' if li['field'] == 2 else 'aperture'
        # solve for imager
        imager_inputs = do_etendue_to_imager(fld_ape_key, etendue_inputs,
                                             etendue_grid, n_0=n_0, n_k=n_k)

    return imager_inputs


def do_field_via_imager(conj_type, imager, etendue_inputs, obj_img_key,
                        etendue_grid, n_0=1, n_k=1):
    input_cell = etendue_inputs['field'][obj_img_key]
    if obj_img_key == 'object':
        output_cell = etendue_grid['field']['image']
        if 'angle' in input_cell:
            efl = imager.f
            obj_ang = input_cell['angle']
            obj_slp = ang2slp(obj_ang)
            output_cell['height'] = efl*obj_slp
        elif 'height' in input_cell:
            m = imager.m
            obj_ht = input_cell['height']
            output_cell['height'] = m*obj_ht

    elif obj_img_key == 'image':
        output_cell = etendue_grid['field']['object']
        if imager.m == 0:  # infinite conjugate
            efl = imager.f
            if 'height' in input_cell:
                img_ht = input_cell['height']
                obj_slp = img_ht/efl
                obj_ang = slp2ang(obj_slp)
                output_cell['angle'] = obj_ang
        else:  # finite conjugate
            m = imager.m
            if 'height' in input_cell:
                img_ht = input_cell['height']
                output_cell['height'] = img_ht/m


def do_aperture_via_imager(conj_type, imager, etendue_inputs, obj_img_key,
                           etendue_grid, n_0=1, n_k=1):
    inpt = etendue_inputs['aperture'][obj_img_key]
    input_cell = etendue_grid['aperture'][obj_img_key]
    if conj_type == 'infinite':
        efl = imager.f

        if obj_img_key == 'object':
            epd = inpt['pupil']
            slpk = (epd/2.0)/efl
            na, fno, pupil = get_aperture_from_slope(imager, slpk, n=n_k)
            output_cell = etendue_grid['aperture']['image']
            output_cell['f/#'] = fno
            output_cell['NA'] = na
            output_cell['pupil'] = pupil

        elif obj_img_key == 'image':
            slpk = get_slope_from_aperture(imager, inpt, n=n_k)
            na, fno, pupil = get_aperture_from_slope(imager, slpk, n=n_k)
            input_cell['f/#'] = fno
            input_cell['NA'] = na
            input_cell['pupil'] = pupil

            epd = 2*slpk*efl

            output_cell = etendue_grid['aperture']['object']
            output_cell['pupil'] = epd

    else:
        mag = imager.m

        if obj_img_key == 'object':
            slp0 = get_slope_from_aperture(imager, inpt, n=n_0)
            na, fno, pupil = get_aperture_from_slope(imager, slp0, n=n_0)
            input_cell['f/#'] = fno
            input_cell['NA'] = na
            input_cell['pupil'] = pupil

            slpk = slp0/mag
            na, fno, pupil = get_aperture_from_slope(imager, slpk, n=n_k)
            output_cell = etendue_grid['aperture']['image']

        elif obj_img_key == 'image':
            slpk = get_slope_from_aperture(imager, inpt, n=n_k)
            na, fno, pupil = get_aperture_from_slope(imager, slpk, n=n_k)
            input_cell['f/#'] = fno
            input_cell['NA'] = na
            input_cell['pupil'] = pupil

            slp0 = mag*slpk
            na, fno, pupil = get_aperture_from_slope(imager, slp0, n=n_0)
            output_cell = etendue_grid['aperture']['object']

        output_cell['f/#'] = fno
        output_cell['NA'] = na
        output_cell['pupil'] = pupil


def do_etendue_to_imager(fld_ape_key, etendue_inputs, etendue_grid,
                         n_0=1.0, n_k=1.0):
    imager_inputs = None
    if fld_ape_key == 'field':
        obj_cell = etendue_inputs['field']['object']
        img_cell = etendue_inputs['field']['image']
        obj_grid = etendue_grid['field']['object']
        img_grid = etendue_grid['field']['image']
        obj_key = obj_cell.keys()
        img_key = img_cell.keys()
        if 'angle' in obj_key:
            obj_ang = obj_cell['angle']
            obj_slp = ang2slp(obj_ang)
            if 'height' in img_key:
                img_ht = img_cell['height']
                img_grid['height'] = img_ht
                efl = img_ht/obj_slp
                imager_inputs = 'f', efl
            else:
                return None
        elif 'height' in obj_key:
            obj_ht = obj_cell['height']
            if 'height' in img_key:
                img_ht = img_cell['height']
                img_grid['height'] = img_ht
                mag = img_ht/obj_ht
                imager_inputs = 'm', mag

    if fld_ape_key == 'aperture':
        obj_cell = etendue_inputs['aperture']['object']
        img_cell = etendue_inputs['aperture']['image']
        obj_grid = etendue_grid['aperture']['object']
        img_grid = etendue_grid['aperture']['image']
        obj_key = obj_cell.keys()
        img_key = img_cell.keys()
        if 'pupil' in obj_key:
            epd = obj_cell['pupil']
            if 'f/#' in img_key:
                fno = img_cell['f/#']
                img_grid['f/#'] = fno
                efl = epd * fno
                imager_inputs = 'f', efl
            if 'NA' in img_key:
                na = img_cell['NA']
                img_grid['NA'] = na
                slpk = na2slp(na, n=n_k)
                efl = (epd/2.0)/slpk
                imager_inputs = 'f', efl
        else:
            if 'NA' in obj_key:
                nao = obj_cell['NA']
                obj_grid['NA'] = nao
                slp0 = na2slp(nao, n=n_0)
            elif 'f/#' in obj_key:
                fno = obj_cell['f/#']
                obj_grid['f/#'] = fno
                slp0 = -1/(2*fno)

            if 'NA' in img_key:
                na = img_cell['NA']
                img_grid['NA'] = na
                slpk = na2slp(na, n=n_k)
            elif 'f/#' in img_key:
                fno = img_cell['f/#']
                img_grid['f/#'] = fno
                slpk = -1/(2*fno)
            mag = slp0/slpk
            imager_inputs = 'm', mag

    return imager_inputs


def fill_in_etendue_data(imager, fld_ape_key, inputs, values, n=1.0):
    if len(inputs) == 0:
        return

    if fld_ape_key == 'field':
        for key in inputs:
            values[key] = inputs[key]

    if fld_ape_key == 'aperture':
        na, fno, pupil = calc_aperture_from_input(imager, inputs, n=n)
        values['NA'] = na
        values['f/#'] = fno
        values['pupil'] = pupil



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
""" manage light and dark interface color schemes

.. Created on Wed Jan 29 20:00:41 2020

.. codeauthor: Michael J. Hayford
"""


solarize = {
    'base03': '#002b36',
    'base02': '#073642',
    'base01': '#586e75',
    'base00': '#657b83',
    'base0': '#839496',
    'base1': '#93a1a1',
    'base2': '#eee8d5',
    'base3': '#fdf6e3',
    'violet': '#6c71c4',
    'blue': '#268bd2',
    'cyan': '#2aa198',
    'green': '#859900',
    'yellow': '#b58900',
    'orange': '#cb4b16',
    'red': '#dc322f',
    'magenta': '#d33682',
    }


def accent_colors(is_dark=True):
    accent = {
        'violet': solarize['violet'],
        'blue': solarize['blue'],
        'cyan': solarize['cyan'],
        'green': solarize['green'],
        'yellow': solarize['yellow'],
        'orange': solarize['orange'],
        'red': solarize['red'],
        'magenta': solarize['magenta'],
        }
    return accent


def foreground_background(is_dark=True):
    if is_dark:
        rgb = {
            'background': solarize['base03'],
            'background1': solarize['base02'],
            'background2': solarize['base01'],
            'foreground': solarize['base0'],
            'foreground1': solarize['base1'],
            'foreground2': solarize['base01']
            }
    else:
        rgb = {
            'background': solarize['base3'],
            'background1': solarize['base2'],
            'background2': solarize['base1'],
            'foreground': solarize['base00'],
            'foreground1': solarize['base01'],
            'foreground2': solarize['base1']
            }
    return rgb
